# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["ExecutionGraphFixup"]

import contextlib
import itertools
from collections import defaultdict
from collections.abc import Sequence
from typing import Any

import networkx as nx
from lsst.pipe.base import QuantumGraph, QuantumNode

from .executionGraphFixup import ExecutionGraphFixup


class ExecFixupDataId(ExecutionGraphFixup):
    """Implementation of ExecutionGraphFixup for ordering of tasks based
    on DataId values.

    This class is a trivial implementation mostly useful as an example,
    though it can be used to make actual fixup instances by defining
    a method that instantiates it, e.g.::

        # lsst/ap/verify/ci_fixup.py

        from lsst.ctrl.mpexec.execFixupDataId import ExecFixupDataId

        def assoc_fixup():
            return ExecFixupDataId(taskLabel="ap_assoc",
                                   dimensions=("visit", "detector"))


    and then executing pipetask::

        pipetask run --graph-fixup=lsst.ap.verify.ci_fixup.assoc_fixup ...

    This will add new dependencies between quanta executed by the task with
    label "ap_assoc". Quanta with higher visit number will depend on quanta
    with lower visit number and their execution will wait until lower visit
    number finishes.

    Parameters
    ----------
    taskLabel : `str`
        The label of the task for which to add dependencies.
    dimensions : `str` or sequence [`str`]
        One or more dimension names, quanta execution will be ordered
        according to values of these dimensions.
    reverse : `bool`, optional
        If `False` (default) then quanta with higher values of dimensions
        will be executed after quanta with lower values, otherwise the order
        is reversed.
    """

    def __init__(self, taskLabel: str, dimensions: str | Sequence[str], reverse: bool = False):
        self.taskLabel = taskLabel
        self.dimensions = dimensions
        self.reverse = reverse
        if isinstance(self.dimensions, str):
            self.dimensions = (self.dimensions,)
        else:
            self.dimensions = tuple(self.dimensions)

    def _key(self, qnode: QuantumNode) -> tuple[Any, ...]:
        """Produce comparison key for quantum data.

        Parameters
        ----------
        qnode : `QuantumNode`
            An individual node in a `~lsst.pipe.base.QuantumGraph`

        Returns
        -------
        key : `tuple`
        """
        dataId = qnode.quantum.dataId
        assert dataId is not None, "Quantum DataId cannot be None"
        key = tuple(dataId[dim] for dim in self.dimensions)
        return key

    def fixupQuanta(self, graph: QuantumGraph) -> QuantumGraph:
        taskDef = graph.findTaskDefByLabel(self.taskLabel)
        if taskDef is None:
            raise ValueError(f"Cannot find task with label {self.taskLabel}")
        quanta = list(graph.getNodesForTask(taskDef))
        keyQuanta = defaultdict(list)
        for q in quanta:
            key = self._key(q)
            keyQuanta[key].append(q)
        keys = sorted(keyQuanta.keys(), reverse=self.reverse)
        networkGraph = graph.graph

        for prev_key, key in itertools.pairwise(keys):
            for prev_node in keyQuanta[prev_key]:
                for node in keyQuanta[key]:
                    # remove any existing edges between the two nodes, but
                    # don't fail if there are not any. Both directions need
                    # tried because in a directed graph, order maters
                    for edge in ((node, prev_node), (prev_node, node)):
                        with contextlib.suppress(nx.NetworkXException):
                            networkGraph.remove_edge(*edge)

                    networkGraph.add_edge(prev_node, node)
        return graph



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from collections.abc import Iterable, Mapping

import click
from lsst.daf.butler.cli.utils import MWOptionDecorator, MWPath, split_commas, unwrap
from lsst.utils.doImport import doImportType

butler_config_option = MWOptionDecorator(
    "-b", "--butler-config", help="Location of the gen3 butler/registry config file."
)


data_query_option = MWOptionDecorator(
    "-d", "--data-query", help="User data selection expression.", metavar="QUERY"
)


debug_option = MWOptionDecorator(
    "--debug", help="Enable debugging output using lsstDebug facility (imports debug.py).", is_flag=True
)

coverage_option = MWOptionDecorator("--coverage", help="Enable coverage output.", is_flag=True)

coverage_report_option = MWOptionDecorator(
    "--cov-report/--no-cov-report",
    help="If coverage is enabled, controls whether to produce an HTML coverage report.",
    default=True,
)

coverage_packages_option = MWOptionDecorator(
    "--cov-packages",
    help=unwrap(
        """Python packages to restrict coverage to.  If none are provided, runs coverage on all packages."""
    ),
    multiple=True,
    callback=split_commas,
)

delete_option = MWOptionDecorator(
    "--delete", callback=split_commas, help="Delete task with given label from pipeline.", multiple=True
)


pdb_option = MWOptionDecorator(
    "--pdb",
    help="Post-mortem debugger to launch for exceptions (defaults to pdb if unspecified; requires a tty).",
    is_flag=False,
    flag_value="pdb",
    default=None,
)


extend_run_option = MWOptionDecorator(
    "--extend-run",
    help=(
        "Instead of creating a new RUN collection, insert datasets into either the one given by "
        "--output-run (if provided) or the first child collection of --output (which must be of type RUN). "
        "This also enables --skip-existing option when building a graph. "
        "When executing a graph this option skips quanta with all existing outputs."
    ),
    is_flag=True,
)


graph_fixup_option = MWOptionDecorator(
    "--graph-fixup",
    help="Name of the class or factory method which makes an instance used for execution graph fixup.",
)


init_only_option = MWOptionDecorator(
    "--init-only",
    help="Do not actually run; just register dataset types and/or save init outputs.",
    is_flag=True,
)


input_option = MWOptionDecorator(
    "-i",
    "--input",
    callback=split_commas,
    default=[],
    help="Comma-separated names of the input collection(s).",
    metavar="COLLECTION",
    multiple=True,
)


rebase_option = MWOptionDecorator(
    "--rebase",
    help=unwrap("""Reset output collection chain if it is inconsistent with --inputs"""),
    is_flag=True,
)


no_versions_option = MWOptionDecorator(
    "--no-versions", help="Do not save or check package versions.", is_flag=True
)


order_pipeline_option = MWOptionDecorator(
    "--order-pipeline",
    help=unwrap(
        """Order tasks in pipeline based on their data
        dependencies, ordering is performed as last step before saving or
        executing pipeline."""
    ),
    is_flag=True,
)


output_option = MWOptionDecorator(
    "-o",
    "--output",
    help=unwrap(
        """Name of the output CHAINED collection. This may either be an
        existing CHAINED collection to use as both input and output
        (incompatible with --input), or a new CHAINED collection created
        to include all inputs (requires --input). In both cases, the
        collection's children will start with an output RUN collection
        that directly holds all new datasets (see --output-run)."""
    ),
    metavar="COLL",
)


output_run_option = MWOptionDecorator(
    "--output-run",
    help=unwrap(
        """Name of the new output RUN collection. If not provided
        then --output must be provided and a new RUN collection will
        be created by appending a timestamp to the value passed with
        --output. If this collection already exists then
        --extend-run must be passed."""
    ),
    metavar="COLL",
)


pipeline_option = MWOptionDecorator(
    "-p",
    "--pipeline",
    help="Location of a pipeline definition file in YAML format.",
    type=MWPath(file_okay=True, dir_okay=False, readable=True),
)


pipeline_dot_option = MWOptionDecorator(
    "--pipeline-dot",
    help="Location for storing GraphViz DOT representation of a pipeline.",
    type=MWPath(writable=True, file_okay=True, dir_okay=False),
)


profile_option = MWOptionDecorator(
    "--profile", help="Dump cProfile statistics to file name.", type=MWPath(file_okay=True, dir_okay=False)
)


prune_replaced_option = MWOptionDecorator(
    "--prune-replaced",
    help=unwrap(
        """Delete the datasets in the collection replaced by
        --replace-run, either just from the datastore
        ('unstore') or by removing them and the RUN completely
        ('purge'). Requires --replace-run."""
    ),
    type=click.Choice(choices=("unstore", "purge"), case_sensitive=False),
)


qgraph_option = MWOptionDecorator(
    "-g",
    "--qgraph",
    help=unwrap(
        """Location for a serialized quantum graph definition (pickle
        file). If this option is given then all input data options and
        pipeline-building options cannot be used.  Can be a URI."""
    ),
)


qgraph_id_option = MWOptionDecorator(
    "--qgraph-id",
    help=unwrap(
        """Quantum graph identifier, if specified must match the
        identifier of the graph loaded from a file. Ignored if graph
        is not loaded from a file."""
    ),
)


qgraph_datastore_records_option = MWOptionDecorator(
    "--qgraph-datastore-records",
    help=unwrap(
        """Include datastore records into generated quantum graph, these records are used by a
        quantum-backed butler.
        """
    ),
    is_flag=True,
)


# I wanted to use default=None here to match Python API but click silently
# replaces None with an empty tuple when multiple=True.
qgraph_node_id_option = MWOptionDecorator(
    "--qgraph-node-id",
    callback=split_commas,
    multiple=True,
    help=unwrap(
        """Only load a specified set of nodes when graph is
        loaded from a file, nodes are identified by UUID
        values. One or more comma-separated integers are
        accepted. By default all nodes are loaded. Ignored if
        graph is not loaded from a file."""
    ),
)

qgraph_header_data_option = MWOptionDecorator(
    "--show-qgraph-header",
    is_flag=True,
    default=False,
    help="Print the headerData for Quantum Graph to the console",
)

qgraph_dot_option = MWOptionDecorator(
    "--qgraph-dot",
    help="Location for storing GraphViz DOT representation of a quantum graph.",
    type=MWPath(writable=True, file_okay=True, dir_okay=False),
)


replace_run_option = MWOptionDecorator(
    "--replace-run",
    help=unwrap(
        """Before creating a new RUN collection in an existing
        CHAINED collection, remove the first child collection
        (which must be of type RUN). This can be used to repeatedly
        write to the same (parent) collection during development,
        but it does not delete the datasets associated with the
        replaced run unless --prune-replaced is also passed.
        Requires --output, and incompatible with --extend-run."""
    ),
    is_flag=True,
)


save_pipeline_option = MWOptionDecorator(
    "-s",
    "--save-pipeline",
    help="Location for storing resulting pipeline definition in YAML format.",
    type=MWPath(dir_okay=False, file_okay=True, writable=True),
)

save_qgraph_option = MWOptionDecorator(
    "-q",
    "--save-qgraph",
    help="URI location for storing a serialized quantum graph definition (pickle file).",
)


save_single_quanta_option = MWOptionDecorator(
    "--save-single-quanta",
    help=unwrap(
        """Format string of locations for storing individual
        quantum graph definition (pickle files). The curly
        brace {} in the input string will be replaced by a
        quantum number. Can be a URI."""
    ),
)


show_option = MWOptionDecorator(
    "--show",
    callback=split_commas,
    help=unwrap(
        """Dump various info to standard output. Possible items are:
        ``config``, ``config=[Task::]<PATTERN>`` or
        ``config=[Task::]<PATTERN>:NOIGNORECASE`` to dump configuration
        fields possibly matching given pattern and/or task label;
        ``history=<FIELD>`` to dump configuration history for a field,
        field name is specified as ``[Task::]<PATTERN>``; ``dump-config``,
        ``dump-config=Task`` to dump complete configuration for a task
        given its label or all tasks; ``pipeline`` to show pipeline
        composition; ``graph`` to show information about quanta;
        ``workflow`` to show information about quanta and their
        dependency; ``tasks`` to show task composition; ``uri`` to show
        predicted dataset URIs of quanta; ``pipeline-graph`` for a
        text-based visualization of the pipeline (tasks and dataset types);
        ``task-graph`` for a text-based visualization of just the tasks.
        With -b, pipeline-graph and task-graph include additional information.
        """
    ),
    metavar="ITEM|ITEM=VALUE",
    multiple=True,
)


skip_existing_in_option = MWOptionDecorator(
    "--skip-existing-in",
    callback=split_commas,
    default=None,
    metavar="COLLECTION",
    multiple=True,
    help=unwrap(
        """If all Quantum outputs already exist in the specified list of
        collections then that Quantum will be excluded from the QuantumGraph.
        """
    ),
)


skip_existing_option = MWOptionDecorator(
    "--skip-existing",
    is_flag=True,
    help=unwrap(
        """This option is equivalent to --skip-existing-in with the name of
        the output RUN collection. If both --skip-existing-in and
        --skip-existing are given then output RUN collection is appended to
        the list of collections."""
    ),
)


clobber_outputs_option = MWOptionDecorator(
    "--clobber-outputs",
    help=(
        "Remove outputs of failed quanta from the output run when they would block the execution of new "
        "quanta with the same data ID (or assume that this will be done, if just building a QuantumGraph).  "
        "Does nothing if --extend-run is not passed."
    ),
    is_flag=True,
)


skip_init_writes_option = MWOptionDecorator(
    "--skip-init-writes",
    help=unwrap(
        """Do not write collection-wide 'init output' datasets
                                                        (e.g.schemas)."""
    ),
    is_flag=True,
)


enable_implicit_threading_option = MWOptionDecorator(
    "--enable-implicit-threading",
    help=unwrap(
        """Do not disable implicit threading use by third-party libraries (e.g. OpenBLAS).
        Implicit threading is always disabled during execution with multiprocessing."""
    ),
    is_flag=True,
)

cores_per_quantum_option = MWOptionDecorator(
    "-n",
    "--cores-per-quantum",
    default=1,
    help=unwrap(
        """Number of cores available to each quantum when executing.
        If '-j' is used each subprocess will be allowed to use this number of cores."""
    ),
    type=click.IntRange(min=1),
)

memory_per_quantum_option = MWOptionDecorator(
    "--memory-per-quantum",
    default="",
    help=unwrap(
        """Memory allocated for each quantum to use when executing.
        This memory allocation is not enforced by the execution system and is purely advisory.
        If '-j' used each subprocess will be allowed to use this amount of memory.
        Units are allowed and the default units for a plain integer are MB.
        For example: '3GB', '3000MB' and '3000' would all result in the same
        memory limit. Default is for no limit."""
    ),
    type=str,
)

task_option = MWOptionDecorator(
    "-t",
    "--task",
    callback=split_commas,
    help=unwrap(
        """Task name to add to pipeline, must be a fully qualified task
        name. Task name can be followed by colon and label name, if label
        is not given then task base name (class name) is used as
        label."""
    ),
    metavar="TASK[:LABEL]",
    multiple=True,
)


timeout_option = MWOptionDecorator(
    "--timeout", type=click.IntRange(min=0), help="Timeout for multiprocessing; maximum wall time (sec)."
)


start_method_option = MWOptionDecorator(
    "--start-method",
    default=None,
    type=click.Choice(choices=["spawn", "fork", "forkserver"]),
    help="Multiprocessing start method, default is platform-specific.",
)


fail_fast_option = MWOptionDecorator(
    "--fail-fast",
    help="Stop processing at first error, default is to process as many tasks as possible.",
    is_flag=True,
)

save_execution_butler_option = MWOptionDecorator(
    "--save-execution-butler",
    help="Export location for an execution-specific butler after making QuantumGraph",
)

mock_option = MWOptionDecorator(
    "--mock",
    help="Mock pipeline execution.",
    is_flag=True,
)

unmocked_dataset_types_option = MWOptionDecorator(
    "--unmocked-dataset-types",
    callback=split_commas,
    default=None,
    metavar="COLLECTION",
    multiple=True,
    help="Names of input dataset types that should not be mocked.",
)


def parse_mock_failure(
    ctx: click.Context, param: click.Option, value: Iterable[str] | None
) -> Mapping[str, tuple[str, type[Exception] | None]]:
    """Parse the --mock-failure option values into the mapping accepted by
    `lsst.pipe.base.tests.mocks.mock_task_defs`.
    """
    result: dict[str, tuple[str, type[Exception] | None]] = {}
    if value is None:
        return result
    for entry in value:
        try:
            task_label, error_type_name, where = entry.split(":", 2)
        except ValueError:
            raise click.UsageError(
                f"Invalid value for --mock-failure option: {entry!r}; "
                "expected a string of the form 'task:error:where'."
            ) from None
        error_type = doImportType(error_type_name) if error_type_name else None
        result[task_label] = (where, error_type)
    return result


mock_failure_option = MWOptionDecorator(
    "--mock-failure",
    callback=parse_mock_failure,
    metavar="LABEL:EXCEPTION:WHERE",
    default=None,
    multiple=True,
    help=unwrap(
        """Specifications for tasks that should be configured to fail
        when mocking execution.  This is a colon-separated 3-tuple, where the
        first entry the task label, the second the fully-qualified exception
        type (empty for ValueError, and the third a string (which typically
        needs to be quoted to be passed as one argument value by the shell) of
        the form passed to --where, indicating which data IDs should fail."""
    ),
)


clobber_execution_butler_option = MWOptionDecorator(
    "--clobber-execution-butler",
    help=unwrap(
        """When creating execution butler overwrite
                                                                   any existing products"""
    ),
    is_flag=True,
)

target_datastore_root_option = MWOptionDecorator(
    "--target-datastore-root",
    help=unwrap(
        """Root directory for datastore of execution butler.
        Default is to use the original datastore.
        """
    ),
)

dataset_query_constraint = MWOptionDecorator(
    "--dataset-query-constraint",
    help=unwrap(
        """When constructing a quantum graph constrain by
        pre-existence of specified dataset types. Valid
        values are `all` for all inputs dataset types in
        pipeline, ``off`` to not consider dataset type
        existence as a constraint, single or comma
        separated list of dataset type names."""
    ),
    default="all",
)

summary_option = MWOptionDecorator(
    "--summary",
    help=(
        "Location for storing job summary (JSON file). Note that the"
        " structure of this file may not be stable."
    ),
    type=MWPath(dir_okay=False, file_okay=True, writable=True),
)


recursive_option = MWOptionDecorator(
    "--recursive",
    is_flag=True,
)

config_search_path_option = MWOptionDecorator(
    "--config-search-path",
    callback=split_commas,
    default=[],
    help="Additional search paths for butler configuration.",
    metavar="PATH",
    multiple=True,
)

update_graph_id_option = MWOptionDecorator(
    "--update-graph-id",
    help=unwrap("Update graph ID with new unique value."),
    is_flag=True,
)

metadata_run_key_option = MWOptionDecorator(
    "--metadata-run-key",
    help=(
        "Quantum graph metadata key for the name of the output run. "
        "Empty string disables update of the metadata. "
        "Default value: output_run."
    ),
    default="output_run",
)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import collections
import contextlib
import re

from lsst.daf.butler.cli.opt import config_file_option, config_option
from lsst.daf.butler.cli.utils import MWCommand, split_commas
from lsst.pipe.base.cli.opt import instrument_option

from .opt import delete_option, task_option

# Class which determines an action that needs to be performed
# when building pipeline, its attributes are:
#   action: the name of the action, e.g. "new_task", "delete_task"
#   label:  task label, can be None if action does not require label
#   value:  argument value excluding task label.
_PipelineAction = collections.namedtuple("_PipelineAction", "action,label,value")


class _PipelineActionType:
    """Class defining a callable type which converts strings into
    ``_PipelineAction`` instances.

    Parameters
    ----------
    action : `str`
        Name of the action, will become `action` attribute of instance.
    regex : `str`
        Regular expression for argument value, it can define groups 'label'
        and 'value' which will become corresponding attributes of a
        returned instance.
    """

    def __init__(self, action: str, regex: str = ".*", valueType: type = str):
        self.action = action
        self.regex = re.compile(regex)
        self.valueType = valueType

    def __call__(self, value: str) -> _PipelineAction:
        match = self.regex.match(value)
        if not match:
            raise TypeError(
                f"Unrecognized syntax for option {self.action!r}: {value!r} "
                f"(does not match pattern {self.regex.pattern})"
            )
        # get "label" group or use None as label
        try:
            label = match.group("label")
        except IndexError:
            label = None
        # if "value" group is not defined use whole string
        with contextlib.suppress(IndexError):
            value = match.group("value")

        value = self.valueType(value)
        return _PipelineAction(self.action, label, value)

    def __repr__(self) -> str:
        """Return string representation of this class."""
        return f"_PipelineActionType(action={self.action})"


_ACTION_ADD_TASK = _PipelineActionType("new_task", "(?P<value>[^:]+)(:(?P<label>.+))?")
_ACTION_DELETE_TASK = _PipelineActionType("delete_task", "(?P<value>)(?P<label>.+)")
_ACTION_CONFIG = _PipelineActionType("config", "(?P<label>.+):(?P<value>.+=.+)")
_ACTION_CONFIG_FILE = _PipelineActionType("configfile", "(?P<label>.+):(?P<value>.+)")
_ACTION_ADD_INSTRUMENT = _PipelineActionType("add_instrument", "(?P<value>[^:]+)")


def makePipelineActions(
    args: list[str],
    taskFlags: list[str] = task_option.opts(),
    deleteFlags: list[str] = delete_option.opts(),
    configFlags: list[str] = config_option.opts(),
    configFileFlags: list[str] = config_file_option.opts(),
    instrumentFlags: list[str] = instrument_option.opts(),
) -> list[_PipelineAction]:
    """Make a list of pipeline actions from a list of option flags and
    values.

    Parameters
    ----------
    args : `list` [`str`]
        The arguments, option flags, and option values in the order they were
        passed in on the command line.
    taskFlags : `list` [`str`], optional
        The option flags to use to recognize a task action, by default
        task_option.opts()
    deleteFlags : `list` [`str`], optional
        The option flags to use to recognize a delete action, by default
        delete_option.opts()
    configFlags : `list` [`str`], optional
        The option flags to use to recognize a config action, by default
        config_option.opts()
    configFileFlags : `list` [`str`], optional
        The option flags to use to recognize a config-file action, by default
        config_file_option.opts()
    instrumentFlags : `list` [`str`], optional
        The option flags to use to recognize an instrument action, by default
        instrument_option.opts()

    Returns
    -------
    pipelineActions : `list` [`_PipelineActionType`]
        A list of pipeline actions constructed form their arguments in args,
        in the order they appeared in args.
    """
    pipelineActions = []
    # iterate up to the second-to-last element, if the second to last element
    # is a key we're looking for, the last item will be its value.
    for i in range(len(args) - 1):
        if args[i] in taskFlags:
            pipelineActions.append(_ACTION_ADD_TASK(args[i + 1]))
        elif args[i] in deleteFlags:
            pipelineActions.append(_ACTION_DELETE_TASK(args[i + 1]))
        elif args[i] in configFlags:
            pipelineActions.append(_ACTION_CONFIG(args[i + 1]))
        elif args[i] in configFileFlags:
            # --config-file allows multiple comma-separated values.
            configfile_args = split_commas(None, None, args[i + 1])
            pipelineActions.extend(_ACTION_CONFIG_FILE(c) for c in configfile_args)
        elif args[i] in instrumentFlags:
            pipelineActions.append(_ACTION_ADD_INSTRUMENT(args[i + 1]))
    return pipelineActions


class PipetaskCommand(MWCommand):
    """Command subclass with pipetask-command specific overrides."""

    extra_epilog = "See 'pipetask --help' for more options."



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simple unit test for cmdLineFwk module.
"""

import contextlib
import logging
import os
import pickle
import re
import shutil
import tempfile
import unittest
from dataclasses import dataclass
from io import StringIO
from types import SimpleNamespace
from typing import NamedTuple

import astropy.units as u
import click
import lsst.pex.config as pexConfig
import lsst.pipe.base.connectionTypes as cT
import lsst.utils.tests
from lsst.ctrl.mpexec import CmdLineFwk, MPGraphExecutorError
from lsst.ctrl.mpexec.cli.opt import run_options
from lsst.ctrl.mpexec.cli.utils import (
    _ACTION_ADD_INSTRUMENT,
    _ACTION_ADD_TASK,
    _ACTION_CONFIG,
    _ACTION_CONFIG_FILE,
    PipetaskCommand,
)
from lsst.ctrl.mpexec.showInfo import ShowInfo
from lsst.daf.butler import (
    CollectionType,
    Config,
    DataCoordinate,
    DatasetRef,
    DimensionConfig,
    DimensionUniverse,
    Quantum,
    Registry,
)
from lsst.daf.butler.core.datasets.type import DatasetType
from lsst.daf.butler.registry import RegistryConfig
from lsst.pipe.base import (
    Instrument,
    Pipeline,
    PipelineTaskConfig,
    PipelineTaskConnections,
    QuantumGraph,
    TaskDef,
)
from lsst.pipe.base.graphBuilder import DatasetQueryConstraintVariant as DQCVariant
from lsst.pipe.base.script import transfer_from_graph
from lsst.pipe.base.tests.simpleQGraph import (
    AddTask,
    AddTaskFactoryMock,
    makeSimpleButler,
    makeSimplePipeline,
    makeSimpleQGraph,
    populateButler,
)
from lsst.utils.tests import temporaryDirectory

logging.basicConfig(level=getattr(logging, os.environ.get("UNIT_TEST_LOGGING_LEVEL", "INFO"), logging.INFO))

# Have to monkey-patch Instrument.fromName() to not retrieve non-existing
# instrument from registry, these tests can run fine without actual instrument
# and implementing full mock for Instrument is too complicated.
Instrument.fromName = lambda name, reg: None


@contextlib.contextmanager
def makeTmpFile(contents=None, suffix=None):
    """Context manager for generating temporary file name.

    Temporary file is deleted on exiting context.

    Parameters
    ----------
    contents : `bytes`
        Data to write into a file.
    """
    fd, tmpname = tempfile.mkstemp(suffix=suffix)
    if contents:
        os.write(fd, contents)
    os.close(fd)
    yield tmpname
    with contextlib.suppress(OSError):
        os.remove(tmpname)


@contextlib.contextmanager
def makeSQLiteRegistry(create=True, universe=None):
    """Context manager to create new empty registry database.

    Yields
    ------
    config : `RegistryConfig`
        Registry configuration for initialized registry database.
    """
    dimensionConfig = universe.dimensionConfig if universe is not None else _makeDimensionConfig()
    with temporaryDirectory() as tmpdir:
        uri = f"sqlite:///{tmpdir}/gen3.sqlite"
        config = RegistryConfig()
        config["db"] = uri
        if create:
            Registry.createFromConfig(config, dimensionConfig=dimensionConfig)
        yield config


class SimpleConnections(PipelineTaskConnections, dimensions=(), defaultTemplates={"template": "simple"}):
    """Test connection class."""

    schema = cT.InitInput(doc="Schema", name="{template}schema", storageClass="SourceCatalog")


class SimpleConfig(PipelineTaskConfig, pipelineConnections=SimpleConnections):
    """Test pipeline config."""

    field = pexConfig.Field(dtype=str, doc="arbitrary string")

    def setDefaults(self):
        PipelineTaskConfig.setDefaults(self)


def _makeArgs(registryConfig=None, **kwargs):
    """Return parsed command line arguments.

    By default butler_config is set to `Config` populated with some defaults,
    it can be overridden completely by keyword argument.

    Parameters
    ----------
    cmd : `str`, optional
        Produce arguments for this pipetask command.
    registryConfig : `RegistryConfig`, optional
        Override for registry configuration.
    **kwargs
        Overrides for other arguments.
    """
    # Use a mock to get the default value of arguments to 'run'.

    mock = unittest.mock.Mock()

    @click.command(cls=PipetaskCommand)
    @run_options()
    def fake_run(ctx, **kwargs):
        """Fake "pipetask run" command for gathering input arguments.

        The arguments & options should always match the arguments & options in
        the "real" command function `lsst.ctrl.mpexec.cli.cmd.run`.
        """
        mock(**kwargs)

    runner = click.testing.CliRunner()
    # --butler-config is the only required option
    result = runner.invoke(fake_run, "--butler-config /")
    if result.exit_code != 0:
        raise RuntimeError(f"Failure getting default args from 'fake_run': {result}")
    mock.assert_called_once()
    args = mock.call_args[1]
    args["enableLsstDebug"] = args.pop("debug")
    args["execution_butler_location"] = args.pop("save_execution_butler")
    if "pipeline_actions" not in args:
        args["pipeline_actions"] = []
    args = SimpleNamespace(**args)

    # override butler_config with our defaults
    if "butler_config" not in kwargs:
        args.butler_config = Config()
        if registryConfig:
            args.butler_config["registry"] = registryConfig
        # The default datastore has a relocatable root, so we need to specify
        # some root here for it to use
        args.butler_config.configFile = "."

    # override arguments from keyword parameters
    for key, value in kwargs.items():
        setattr(args, key, value)
    args.dataset_query_constraint = DQCVariant.fromExpression(args.dataset_query_constraint)
    return args


class FakeDSType(NamedTuple):
    """A fake `~lsst.daf.butler.DatasetType` class used for testing."""

    name: str


@dataclass(frozen=True)
class FakeDSRef:
    """A fake `~lsst.daf.butler.DatasetRef` class used for testing."""

    datasetType: str
    dataId: tuple

    def isComponent(self):
        return False


# Task class name used by tests, needs to be importable
_TASK_CLASS = "lsst.pipe.base.tests.simpleQGraph.AddTask"


def _makeDimensionConfig():
    """Make a simple dimension universe configuration."""
    return DimensionConfig(
        {
            "version": 1,
            "namespace": "ctrl_mpexec_test",
            "skypix": {
                "common": "htm7",
                "htm": {
                    "class": "lsst.sphgeom.HtmPixelization",
                    "max_level": 24,
                },
            },
            "elements": {
                "A": {
                    "keys": [
                        {
                            "name": "id",
                            "type": "int",
                        }
                    ],
                    "storage": {
                        "cls": "lsst.daf.butler.registry.dimensions.table.TableDimensionRecordStorage",
                    },
                },
                "B": {
                    "keys": [
                        {
                            "name": "id",
                            "type": "int",
                        }
                    ],
                    "storage": {
                        "cls": "lsst.daf.butler.registry.dimensions.table.TableDimensionRecordStorage",
                    },
                },
            },
            "packers": {},
        }
    )


def _makeQGraph():
    """Make a trivial QuantumGraph with one quantum.

    The only thing that we need to do with this quantum graph is to pickle
    it, the quanta in this graph are not usable for anything else.

    Returns
    -------
    qgraph : `~lsst.pipe.base.QuantumGraph`
    """
    universe = DimensionUniverse(config=_makeDimensionConfig())
    fakeDSType = DatasetType("A", (), storageClass="ExposureF", universe=universe)
    taskDef = TaskDef(taskName=_TASK_CLASS, config=AddTask.ConfigClass(), taskClass=AddTask)
    quanta = [
        Quantum(
            taskName=_TASK_CLASS,
            inputs={
                fakeDSType: [
                    DatasetRef(
                        fakeDSType,
                        DataCoordinate.standardize({"A": 1, "B": 2}, universe=universe),
                        run="fake_run",
                    )
                ]
            },
        )
    ]  # type: ignore
    qgraph = QuantumGraph({taskDef: set(quanta)}, universe=universe)
    return qgraph


class CmdLineFwkTestCase(unittest.TestCase):
    """A test case for CmdLineFwk"""

    def testMakePipeline(self):
        """Tests for CmdLineFwk.makePipeline method"""
        fwk = CmdLineFwk()

        # make empty pipeline
        args = _makeArgs()
        pipeline = fwk.makePipeline(args)
        self.assertIsInstance(pipeline, Pipeline)
        self.assertEqual(len(pipeline), 0)

        # few tests with serialization
        with makeTmpFile() as tmpname:
            # make empty pipeline and store it in a file
            args = _makeArgs(save_pipeline=tmpname)
            pipeline = fwk.makePipeline(args)
            self.assertIsInstance(pipeline, Pipeline)

            # read pipeline from a file
            args = _makeArgs(pipeline=tmpname)
            pipeline = fwk.makePipeline(args)
            self.assertIsInstance(pipeline, Pipeline)
            self.assertEqual(len(pipeline), 0)

        # single task pipeline, task name can be anything here
        actions = [_ACTION_ADD_TASK("TaskOne:task1")]
        args = _makeArgs(pipeline_actions=actions)
        pipeline = fwk.makePipeline(args)
        self.assertIsInstance(pipeline, Pipeline)
        self.assertEqual(len(pipeline), 1)

        # many task pipeline
        actions = [
            _ACTION_ADD_TASK("TaskOne:task1a"),
            _ACTION_ADD_TASK("TaskTwo:task2"),
            _ACTION_ADD_TASK("TaskOne:task1b"),
        ]
        args = _makeArgs(pipeline_actions=actions)
        pipeline = fwk.makePipeline(args)
        self.assertIsInstance(pipeline, Pipeline)
        self.assertEqual(len(pipeline), 3)

        # single task pipeline with config overrides, need real task class
        actions = [_ACTION_ADD_TASK(f"{_TASK_CLASS}:task"), _ACTION_CONFIG("task:addend=100")]
        args = _makeArgs(pipeline_actions=actions)
        pipeline = fwk.makePipeline(args)
        taskDefs = list(pipeline.toExpandedPipeline())
        self.assertEqual(len(taskDefs), 1)
        self.assertEqual(taskDefs[0].config.addend, 100)

        overrides = b"config.addend = 1000\n"
        with makeTmpFile(overrides) as tmpname:
            actions = [_ACTION_ADD_TASK(f"{_TASK_CLASS}:task"), _ACTION_CONFIG_FILE("task:" + tmpname)]
            args = _makeArgs(pipeline_actions=actions)
            pipeline = fwk.makePipeline(args)
            taskDefs = list(pipeline.toExpandedPipeline())
            self.assertEqual(len(taskDefs), 1)
            self.assertEqual(taskDefs[0].config.addend, 1000)

        # Check --instrument option, for now it only checks that it does not
        # crash.
        actions = [_ACTION_ADD_TASK(f"{_TASK_CLASS}:task"), _ACTION_ADD_INSTRUMENT("Instrument")]
        args = _makeArgs(pipeline_actions=actions)
        pipeline = fwk.makePipeline(args)

    def testMakeGraphFromSave(self):
        """Tests for CmdLineFwk.makeGraph method.

        Only most trivial case is tested that does not do actual graph
        building.
        """
        fwk = CmdLineFwk()

        with makeTmpFile(suffix=".qgraph") as tmpname, makeSQLiteRegistry() as registryConfig:
            # make non-empty graph and store it in a file
            qgraph = _makeQGraph()
            with open(tmpname, "wb") as saveFile:
                qgraph.save(saveFile)
            args = _makeArgs(qgraph=tmpname, registryConfig=registryConfig, execution_butler_location=None)
            qgraph = fwk.makeGraph(None, args)
            self.assertIsInstance(qgraph, QuantumGraph)
            self.assertEqual(len(qgraph), 1)

            # will fail if graph id does not match
            args = _makeArgs(
                qgraph=tmpname,
                qgraph_id="R2-D2 is that you?",
                registryConfig=registryConfig,
                execution_butler_location=None,
            )
            with self.assertRaisesRegex(ValueError, "graphID does not match"):
                fwk.makeGraph(None, args)

            # save with wrong object type
            with open(tmpname, "wb") as saveFile:
                pickle.dump({}, saveFile)
            args = _makeArgs(qgraph=tmpname, registryConfig=registryConfig, execution_butler_location=None)
            with self.assertRaises(ValueError):
                fwk.makeGraph(None, args)

            # reading empty graph from pickle should work but makeGraph()
            # will return None.
            qgraph = QuantumGraph({}, universe=DimensionUniverse(_makeDimensionConfig()))
            with open(tmpname, "wb") as saveFile:
                qgraph.save(saveFile)
            args = _makeArgs(qgraph=tmpname, registryConfig=registryConfig, execution_butler_location=None)
            qgraph = fwk.makeGraph(None, args)
            self.assertIs(qgraph, None)

    def testShowPipeline(self):
        """Test for --show options for pipeline."""
        fwk = CmdLineFwk()

        actions = [_ACTION_ADD_TASK(f"{_TASK_CLASS}:task"), _ACTION_CONFIG("task:addend=100")]
        args = _makeArgs(pipeline_actions=actions)
        pipeline = fwk.makePipeline(args)

        with self.assertRaises(ValueError):
            ShowInfo(["unrecognized", "config"])

        stream = StringIO()
        show = ShowInfo(
            ["pipeline", "config", "history=task::addend", "tasks", "dump-config", "config=task::add*"],
            stream=stream,
        )
        show.show_pipeline_info(pipeline, None)
        self.assertEqual(show.unhandled, frozenset({}))
        stream.seek(0)
        output = stream.read()
        self.assertIn("config.addend=100", output)  # config option
        self.assertIn("addend\n3", output)  # History output
        self.assertIn("class: lsst.pipe.base.tests.simpleQGraph.AddTask", output)  # pipeline

        show = ShowInfo(["pipeline", "uri"], stream=stream)
        show.show_pipeline_info(pipeline, None)
        self.assertEqual(show.unhandled, frozenset({"uri"}))
        self.assertEqual(show.handled, {"pipeline"})

        stream = StringIO()
        show = ShowInfo(["config=task::addend.missing"], stream=stream)  # No match
        show.show_pipeline_info(pipeline, None)
        stream.seek(0)
        output = stream.read().strip()
        self.assertEqual("### Configuration for task `task'", output)

        stream = StringIO()
        show = ShowInfo(["config=task::addEnd:NOIGNORECASE"], stream=stream)  # No match
        show.show_pipeline_info(pipeline, None)
        stream.seek(0)
        output = stream.read().strip()
        self.assertEqual("### Configuration for task `task'", output)

        stream = StringIO()
        show = ShowInfo(["pipeline-graph"], stream=stream)  # No match
        show.show_pipeline_info(pipeline, None)
        stream.seek(0)
        output = stream.read().strip()
        self.assertEqual(
            "\n".join(
                [
                    "○  add_dataset_in",
                    "│",
                    "■  task",
                    "│",
                    "◍  add_dataset_out, add2_dataset_out",
                ]
            ),
            output,
        )

        stream = StringIO()
        show = ShowInfo(["task-graph"], stream=stream)  # No match
        show.show_pipeline_info(pipeline, None)
        stream.seek(0)
        output = stream.read().strip()
        self.assertEqual("■  task", output)

        stream = StringIO()
        show = ShowInfo(["config=task::addEnd"], stream=stream)  # Match but warns
        show.show_pipeline_info(pipeline, None)
        stream.seek(0)
        output = stream.read().strip()
        self.assertIn("NOIGNORECASE", output)

        show = ShowInfo(["dump-config=notask"])
        with self.assertRaises(ValueError) as cm:
            show.show_pipeline_info(pipeline, None)
        self.assertIn("Pipeline has no tasks named notask", str(cm.exception))

        show = ShowInfo(["history"])
        with self.assertRaises(ValueError) as cm:
            show.show_pipeline_info(pipeline, None)
        self.assertIn("Please provide a value", str(cm.exception))

        show = ShowInfo(["history=notask::param"])
        with self.assertRaises(ValueError) as cm:
            show.show_pipeline_info(pipeline, None)
        self.assertIn("Pipeline has no tasks named notask", str(cm.exception))

    def test_execution_resources_parameters(self) -> None:
        """Test creation of the ExecutionResources from command line."""
        fwk = CmdLineFwk()

        for params, num_cores, max_mem in (
            ((None, None), 1, None),
            ((5, ""), 5, None),
            ((None, "50"), 1, 50 * u.MB),
            ((5, "50 GB"), 5, 50 * u.GB),
        ):
            kwargs = {}
            for k, v in zip(("cores_per_quantum", "memory_per_quantum"), params, strict=True):
                if v is not None:
                    kwargs[k] = v
            args = _makeArgs(**kwargs)
            res = fwk._make_execution_resources(args)
            self.assertEqual(res.num_cores, num_cores)
            self.assertEqual(res.max_mem, max_mem)

        args = _makeArgs(memory_per_quantum="50m")
        with self.assertRaises(u.UnitConversionError):
            fwk._make_execution_resources(args)


class CmdLineFwkTestCaseWithButler(unittest.TestCase):
    """A test case for CmdLineFwk"""

    def setUp(self):
        super().setUpClass()
        self.root = tempfile.mkdtemp()
        self.nQuanta = 5
        self.pipeline = makeSimplePipeline(nQuanta=self.nQuanta)

    def tearDown(self):
        shutil.rmtree(self.root, ignore_errors=True)
        super().tearDownClass()

    def testSimpleQGraph(self):
        """Test successfull execution of trivial quantum graph."""
        args = _makeArgs(butler_config=self.root, input="test", output="output")
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock()

        qgraph = fwk.makeGraph(self.pipeline, args)
        self.assertEqual(len(qgraph.taskGraph), self.nQuanta)
        self.assertEqual(len(qgraph), self.nQuanta)

        # Ensure that the output run used in the graph is also used in
        # the pipeline execution. It is possible for makeGraph and runPipeline
        # to calculate time-stamped runs across a second boundary.
        args.output_run = qgraph.metadata["output_run"]

        # run whole thing
        fwk.runPipeline(qgraph, taskFactory, args)
        self.assertEqual(taskFactory.countExec, self.nQuanta)

        # test that we've disabled implicit threading
        self.assertEqual(os.environ["OMP_NUM_THREADS"], "1")

    def testSimpleQGraph_rebase(self):
        """Test successful execution of trivial quantum graph, with --rebase
        used to force redefinition of the output collection.
        """
        # Pass one input collection here for the usual test setup; we'll
        # override it later.
        args = _makeArgs(butler_config=self.root, input="test1", output="output")
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock()

        # We'll actually pass two input collections in.  One is empty, but
        # the stuff we're testing here doesn't care.
        args.input = ["test2", "test1"]
        butler.registry.registerCollection("test2", CollectionType.RUN)

        # Set up the output collection with a sequence that doesn't end the
        # same way as the input collection.  This is normally an error.
        butler.registry.registerCollection("output", CollectionType.CHAINED)
        butler.registry.registerCollection("unexpected_input", CollectionType.RUN)
        butler.registry.registerCollection("output/run0", CollectionType.RUN)
        butler.registry.setCollectionChain("output", ["test2", "unexpected_input", "test1", "output/run0"])

        # Without --rebase, the inconsistent input and output collections are
        # an error.
        with self.assertRaises(ValueError):
            fwk.makeGraph(self.pipeline, args)

        # With --rebase, the output collection gets redefined.
        args.rebase = True
        qgraph = fwk.makeGraph(self.pipeline, args)

        self.assertEqual(len(qgraph.taskGraph), self.nQuanta)
        self.assertEqual(len(qgraph), self.nQuanta)

        # Ensure that the output run used in the graph is also used in
        # the pipeline execution. It is possible for makeGraph and runPipeline
        # to calculate time-stamped runs across a second boundary.
        args.output_run = qgraph.metadata["output_run"]

        fwk.runPipeline(qgraph, taskFactory, args)
        self.assertEqual(taskFactory.countExec, self.nQuanta)

        butler.registry.refresh()
        self.assertEqual(
            list(butler.registry.getCollectionChain("output")),
            [args.output_run, "output/run0", "test2", "test1", "unexpected_input"],
        )

    def test_simple_qgraph_qbb(self):
        """Test successful execution of trivial quantum graph in QBB mode."""
        args = _makeArgs(
            butler_config=self.root, input="test", output="output", qgraph_datastore_records=True
        )
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock()

        qgraph = fwk.makeGraph(self.pipeline, args)
        self.assertEqual(len(qgraph.taskGraph), self.nQuanta)
        self.assertEqual(len(qgraph), self.nQuanta)

        # Ensure that the output run used in the graph is also used in
        # the pipeline execution. It is possible for makeGraph and runPipeline
        # to calculate time-stamped runs across a second boundary.
        output_run = qgraph.metadata["output_run"]
        args.output_run = output_run

        # QBB must run from serialized graph.
        with tempfile.NamedTemporaryFile(suffix=".qgraph") as temp_graph:
            qgraph.saveUri(temp_graph.name)

            args = _makeArgs(butler_config=self.root, qgraph=temp_graph.name, config_search_path=[])

            # Check that pre-exec-init can run.
            fwk.preExecInitQBB(taskFactory, args)

            # Run whole thing.
            fwk.runGraphQBB(taskFactory, args)

            # Transfer the datasets to the butler.
            n1 = transfer_from_graph(temp_graph.name, self.root, True, False, False)
            self.assertEqual(n1, 31)

        self.assertEqual(taskFactory.countExec, self.nQuanta)

        # Update the output run and try again.
        new_output_run = output_run + "_new"
        qgraph.updateRun(new_output_run, metadata_key="output_run", update_graph_id=True)
        self.assertEqual(qgraph.metadata["output_run"], new_output_run)

        taskFactory = AddTaskFactoryMock()
        with tempfile.NamedTemporaryFile(suffix=".qgraph") as temp_graph:
            qgraph.saveUri(temp_graph.name)

            args = _makeArgs(butler_config=self.root, qgraph=temp_graph.name, config_search_path=[])

            # Check that pre-exec-init can run.
            fwk.preExecInitQBB(taskFactory, args)

            # Run whole thing.
            fwk.runGraphQBB(taskFactory, args)

            # Transfer the datasets to the butler.
            n2 = transfer_from_graph(temp_graph.name, self.root, True, False, False)
            self.assertEqual(n1, n2)

    def testEmptyQGraph(self):
        """Test that making an empty QG produces the right error messages."""
        # We make QG generation fail by populating one input collection in the
        # butler while using a different one (that we only register, not
        # populate) to make the QG.
        args = _makeArgs(butler_config=self.root, input="bad_input", output="output")
        butler = makeSimpleButler(self.root, run="good_input", inMemory=False)
        butler.registry.registerCollection("bad_input")
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        with self.assertLogs(level=logging.CRITICAL) as cm:
            qgraph = fwk.makeGraph(self.pipeline, args)
        self.assertRegex(
            cm.output[0], ".*Initial data ID query returned no rows, so QuantumGraph will be empty.*"
        )
        self.assertRegex(cm.output[1], ".*No datasets.*bad_input.*")
        self.assertIsNone(qgraph)

    def testSimpleQGraphNoSkipExisting_inputs(self):
        """Test for case when output data for one task already appears in
        _input_ collection, but no ``--extend-run`` or ``-skip-existing``
        option is present.
        """
        args = _makeArgs(
            butler_config=self.root,
            input="test",
            output="output",
        )
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(
            self.pipeline,
            butler,
            datasetTypes={
                args.input: [
                    "add_dataset0",
                    "add_dataset1",
                    "add2_dataset1",
                    "add_init_output1",
                    "task0_config",
                    "task0_metadata",
                    "task0_log",
                ]
            },
        )

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock()

        qgraph = fwk.makeGraph(self.pipeline, args)
        self.assertEqual(len(qgraph.taskGraph), self.nQuanta)
        # With current implementation graph has all nQuanta quanta, but when
        # executing one quantum is skipped.
        self.assertEqual(len(qgraph), self.nQuanta)

        # Ensure that the output run used in the graph is also used in
        # the pipeline execution. It is possible for makeGraph and runPipeline
        # to calculate time-stamped runs across a second boundary.
        args.output_run = qgraph.metadata["output_run"]

        # run whole thing
        fwk.runPipeline(qgraph, taskFactory, args)
        self.assertEqual(taskFactory.countExec, self.nQuanta)

    def testSimpleQGraphSkipExisting_inputs(self):
        """Test for ``--skip-existing`` with output data for one task already
        appears in _input_ collection. No ``--extend-run`` option is needed
        for this case.
        """
        args = _makeArgs(
            butler_config=self.root,
            input="test",
            output="output",
            skip_existing_in=("test",),
        )
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(
            self.pipeline,
            butler,
            datasetTypes={
                args.input: [
                    "add_dataset0",
                    "add_dataset1",
                    "add2_dataset1",
                    "add_init_output1",
                    "task0_config",
                    "task0_metadata",
                    "task0_log",
                ]
            },
        )

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock()

        qgraph = fwk.makeGraph(self.pipeline, args)
        # If all quanta are skipped, the task is not included in the graph.
        self.assertEqual(len(qgraph.taskGraph), self.nQuanta - 1)
        self.assertEqual(len(qgraph), self.nQuanta - 1)

        # Ensure that the output run used in the graph is also used in
        # the pipeline execution. It is possible for makeGraph and runPipeline
        # to calculate time-stamped runs across a second boundary.
        args.output_run = qgraph.metadata["output_run"]

        # run whole thing
        fwk.runPipeline(qgraph, taskFactory, args)
        self.assertEqual(taskFactory.countExec, self.nQuanta - 1)

    def testSimpleQGraphSkipExisting_outputs(self):
        """Test for ``--skip-existing`` with output data for one task already
        appears in _output_ collection. The ``--extend-run`` option is needed
        for this case.
        """
        args = _makeArgs(
            butler_config=self.root,
            input="test",
            output_run="output/run",
            skip_existing_in=("output/run",),
        )
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(
            self.pipeline,
            butler,
            datasetTypes={
                args.input: ["add_dataset0"],
                args.output_run: [
                    "add_dataset1",
                    "add2_dataset1",
                    "add_init_output1",
                    "task0_metadata",
                    "task0_log",
                    "task0_config",
                ],
            },
        )

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock()

        # fails without --extend-run
        with self.assertRaisesRegex(ValueError, "--extend-run was not given"):
            qgraph = fwk.makeGraph(self.pipeline, args)

        # retry with --extend-run
        args.extend_run = True
        qgraph = fwk.makeGraph(self.pipeline, args)

        # First task has no remaining quanta, so is left out completely.
        self.assertEqual(len(qgraph.taskGraph), self.nQuanta - 1)
        # Graph does not include quantum for first task.
        self.assertEqual(len(qgraph), self.nQuanta - 1)

        # run whole thing
        fwk.runPipeline(qgraph, taskFactory, args)
        self.assertEqual(taskFactory.countExec, self.nQuanta - 1)

    def testSimpleQGraphOutputsFail(self):
        """Test continuing execution of trivial quantum graph with partial
        outputs.
        """
        args = _makeArgs(butler_config=self.root, input="test", output="output")
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock(stopAt=3)

        qgraph = fwk.makeGraph(self.pipeline, args)
        self.assertEqual(len(qgraph), self.nQuanta)

        # Ensure that the output run used in the graph is also used in
        # the pipeline execution. It is possible for makeGraph and runPipeline
        # to calculate time-stamped runs across a second boundary.
        args.output_run = qgraph.metadata["output_run"]

        # run first three quanta
        with self.assertRaises(MPGraphExecutorError):
            fwk.runPipeline(qgraph, taskFactory, args)
        self.assertEqual(taskFactory.countExec, 3)

        butler.registry.refresh()

        # drop one of the two outputs from one task
        ref1 = butler.registry.findDataset(
            "add2_dataset2", collections=args.output, instrument="INSTR", detector=0
        )
        self.assertIsNotNone(ref1)
        # also drop the metadata output
        ref2 = butler.registry.findDataset(
            "task1_metadata", collections=args.output, instrument="INSTR", detector=0
        )
        self.assertIsNotNone(ref2)
        butler.pruneDatasets([ref1, ref2], disassociate=True, unstore=True, purge=True)

        # Ensure that the output run used in the graph is also used in
        # the pipeline execution. It is possible for makeGraph and runPipeline
        # to calculate time-stamped runs across a second boundary.
        args.output_run = qgraph.metadata["output_run"]

        taskFactory.stopAt = -1
        args.skip_existing_in = (args.output,)
        args.extend_run = True
        args.no_versions = True
        with self.assertRaises(MPGraphExecutorError):
            fwk.runPipeline(qgraph, taskFactory, args)

    def testSimpleQGraphClobberOutputs(self):
        """Test continuing execution of trivial quantum graph with
        --clobber-outputs.
        """
        args = _makeArgs(butler_config=self.root, input="test", output="output")
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock(stopAt=3)

        qgraph = fwk.makeGraph(self.pipeline, args)

        # should have one task and number of quanta
        self.assertEqual(len(qgraph), self.nQuanta)

        # Ensure that the output run used in the graph is also used in
        # the pipeline execution. It is possible for makeGraph and runPipeline
        # to calculate time-stamped runs across a second boundary.
        args.output_run = qgraph.metadata["output_run"]

        # run first three quanta
        with self.assertRaises(MPGraphExecutorError):
            fwk.runPipeline(qgraph, taskFactory, args)
        self.assertEqual(taskFactory.countExec, 3)

        butler.registry.refresh()

        # drop one of the two outputs from one task
        ref1 = butler.registry.findDataset(
            "add2_dataset2", collections=args.output, dataId=dict(instrument="INSTR", detector=0)
        )
        self.assertIsNotNone(ref1)
        # also drop the metadata output
        ref2 = butler.registry.findDataset(
            "task1_metadata", collections=args.output, dataId=dict(instrument="INSTR", detector=0)
        )
        self.assertIsNotNone(ref2)
        butler.pruneDatasets([ref1, ref2], disassociate=True, unstore=True, purge=True)

        taskFactory.stopAt = -1
        args.skip_existing = True
        args.extend_run = True
        args.clobber_outputs = True
        args.no_versions = True
        fwk.runPipeline(qgraph, taskFactory, args)
        # number of executed quanta is incremented
        self.assertEqual(taskFactory.countExec, self.nQuanta + 1)

    def testSimpleQGraphReplaceRun(self):
        """Test repeated execution of trivial quantum graph with
        --replace-run.
        """
        args = _makeArgs(butler_config=self.root, input="test", output="output", output_run="output/run1")
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        taskFactory = AddTaskFactoryMock()

        qgraph = fwk.makeGraph(self.pipeline, args)

        # should have one task and number of quanta
        self.assertEqual(len(qgraph), self.nQuanta)

        # deep copy is needed because quanta are updated in place
        fwk.runPipeline(qgraph, taskFactory, args)
        self.assertEqual(taskFactory.countExec, self.nQuanta)

        # need to refresh collections explicitly (or make new butler/registry)
        butler.registry.refresh()
        collections = set(butler.registry.queryCollections(...))
        self.assertEqual(collections, {"test", "output", "output/run1"})

        # number of datasets written by pipeline:
        #  - nQuanta of init_outputs
        #  - nQuanta of configs
        #  - packages (single dataset)
        #  - nQuanta * two output datasets
        #  - nQuanta of metadata
        #  - nQuanta of log output
        n_outputs = self.nQuanta * 6 + 1
        refs = butler.registry.queryDatasets(..., collections="output/run1")
        self.assertEqual(len(list(refs)), n_outputs)

        # re-run with --replace-run (--inputs is ignored, as long as it hasn't
        # changed)
        args.replace_run = True
        args.output_run = "output/run2"
        qgraph = fwk.makeGraph(self.pipeline, args)
        fwk.runPipeline(qgraph, taskFactory, args)

        butler.registry.refresh()
        collections = set(butler.registry.queryCollections(...))
        self.assertEqual(collections, {"test", "output", "output/run1", "output/run2"})

        # new output collection
        refs = butler.registry.queryDatasets(..., collections="output/run2")
        self.assertEqual(len(list(refs)), n_outputs)

        # old output collection is still there
        refs = butler.registry.queryDatasets(..., collections="output/run1")
        self.assertEqual(len(list(refs)), n_outputs)

        # re-run with --replace-run and --prune-replaced=unstore
        args.replace_run = True
        args.prune_replaced = "unstore"
        args.output_run = "output/run3"
        qgraph = fwk.makeGraph(self.pipeline, args)
        fwk.runPipeline(qgraph, taskFactory, args)

        butler.registry.refresh()
        collections = set(butler.registry.queryCollections(...))
        self.assertEqual(collections, {"test", "output", "output/run1", "output/run2", "output/run3"})

        # new output collection
        refs = butler.registry.queryDatasets(..., collections="output/run3")
        self.assertEqual(len(list(refs)), n_outputs)

        # old output collection is still there, and it has all datasets but
        # non-InitOutputs are not in datastore
        refs = butler.registry.queryDatasets(..., collections="output/run2")
        refs = list(refs)
        self.assertEqual(len(refs), n_outputs)
        initOutNameRe = re.compile("packages|task.*_config|add_init_output.*")
        for ref in refs:
            if initOutNameRe.fullmatch(ref.datasetType.name):
                butler.get(ref)
            else:
                with self.assertRaises(FileNotFoundError):
                    butler.get(ref)

        # re-run with --replace-run and --prune-replaced=purge
        # This time also remove --input; passing the same inputs that we
        # started with and not passing inputs at all should be equivalent.
        args.input = None
        args.replace_run = True
        args.prune_replaced = "purge"
        args.output_run = "output/run4"
        qgraph = fwk.makeGraph(self.pipeline, args)
        fwk.runPipeline(qgraph, taskFactory, args)

        butler.registry.refresh()
        collections = set(butler.registry.queryCollections(...))
        # output/run3 should disappear now
        self.assertEqual(collections, {"test", "output", "output/run1", "output/run2", "output/run4"})

        # new output collection
        refs = butler.registry.queryDatasets(..., collections="output/run4")
        self.assertEqual(len(list(refs)), n_outputs)

        # Trying to run again with inputs that aren't exactly what we started
        # with is an error, and the kind that should not modify the data repo.
        with self.assertRaises(ValueError):
            args.input = ["test", "output/run2"]
            args.prune_replaced = None
            args.replace_run = True
            args.output_run = "output/run5"
            qgraph = fwk.makeGraph(self.pipeline, args)
            fwk.runPipeline(qgraph, taskFactory, args)
        butler.registry.refresh()
        collections = set(butler.registry.queryCollections(...))
        self.assertEqual(collections, {"test", "output", "output/run1", "output/run2", "output/run4"})
        with self.assertRaises(ValueError):
            args.input = ["output/run2", "test"]
            args.prune_replaced = None
            args.replace_run = True
            args.output_run = "output/run6"
            qgraph = fwk.makeGraph(self.pipeline, args)
            fwk.runPipeline(qgraph, taskFactory, args)
        butler.registry.refresh()
        collections = set(butler.registry.queryCollections(...))
        self.assertEqual(collections, {"test", "output", "output/run1", "output/run2", "output/run4"})

    def testSubgraph(self):
        """Test successful execution of trivial quantum graph."""
        args = _makeArgs(butler_config=self.root, input="test", output="output")
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        qgraph = fwk.makeGraph(self.pipeline, args)

        # Select first two nodes for execution. This depends on node ordering
        # which I assume is the same as execution order.
        nNodes = 2
        nodeIds = [node.nodeId for node in qgraph]
        nodeIds = nodeIds[:nNodes]

        self.assertEqual(len(qgraph.taskGraph), self.nQuanta)
        self.assertEqual(len(qgraph), self.nQuanta)

        with (
            makeTmpFile(suffix=".qgraph") as tmpname,
            makeSQLiteRegistry(universe=butler.dimensions) as registryConfig,
        ):
            with open(tmpname, "wb") as saveFile:
                qgraph.save(saveFile)

            args = _makeArgs(
                qgraph=tmpname,
                qgraph_node_id=nodeIds,
                registryConfig=registryConfig,
                execution_butler_location=None,
            )
            fwk = CmdLineFwk()

            # load graph, should only read a subset
            qgraph = fwk.makeGraph(pipeline=None, args=args)
            self.assertEqual(len(qgraph), nNodes)

    def testShowGraph(self):
        """Test for --show options for quantum graph."""
        nQuanta = 2
        butler, qgraph = makeSimpleQGraph(nQuanta, root=self.root)

        show = ShowInfo(["graph"])
        show.show_graph_info(qgraph)
        self.assertEqual(show.handled, {"graph"})

    def testShowGraphWorkflow(self):
        nQuanta = 2
        butler, qgraph = makeSimpleQGraph(nQuanta, root=self.root)

        show = ShowInfo(["workflow"])
        show.show_graph_info(qgraph)
        self.assertEqual(show.handled, {"workflow"})

        # TODO: cannot test "uri" option presently, it instantiates
        # butler from command line options and there is no way to pass butler
        # mock to that code.
        show = ShowInfo(["uri"])
        with self.assertRaises(ValueError):  # No args given
            show.show_graph_info(qgraph)

    def testSimpleQGraphDatastoreRecords(self):
        """Test quantum graph generation with --qgraph-datastore-records."""
        args = _makeArgs(
            butler_config=self.root, input="test", output="output", qgraph_datastore_records=True
        )
        butler = makeSimpleButler(self.root, run=args.input, inMemory=False)
        populateButler(self.pipeline, butler)

        fwk = CmdLineFwk()
        qgraph = fwk.makeGraph(self.pipeline, args)
        self.assertEqual(len(qgraph), self.nQuanta)
        for i, qnode in enumerate(qgraph):
            quantum = qnode.quantum
            self.assertIsNotNone(quantum.datastore_records)
            # only the first quantum has a pre-existing input
            if i == 0:
                datastore_name = "FileDatastore@<butlerRoot>"
                self.assertEqual(set(quantum.datastore_records.keys()), {datastore_name})
                records_data = quantum.datastore_records[datastore_name]
                records = dict(records_data.records)
                self.assertEqual(len(records), 1)
                _, records = records.popitem()
                records = records["file_datastore_records"]
                self.assertEqual(
                    [record.path for record in records],
                    ["test/add_dataset0/add_dataset0_INSTR_det0_test.pickle"],
                )
            else:
                self.assertEqual(quantum.datastore_records, {})


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """General file leak detection."""


def setup_module(module):
    """Initialize pytest module."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Few utility methods used by the rest of a package.
"""

__all__ = ["profile", "printTable", "filterTasks", "subTaskIter"]

# -------------------------------
#  Imports of standard modules --
# -------------------------------
import contextlib
import logging
from collections.abc import Iterator

# -----------------------------
#  Imports for other modules --
# -----------------------------
import lsst.pex.config as pexConfig
from lsst.pipe.base import Pipeline, TaskDef

# ----------------------------------
#  Local non-exported definitions --
# ----------------------------------

# ------------------------
#  Exported definitions --
# ------------------------


@contextlib.contextmanager
def profile(filename: str, log: logging.Logger | None = None) -> Iterator:
    """Context manager for profiling with cProfile

    Parameters
    ----------
    filename :
        Filename to which to write profile (profiling disabled if `None`
        or empty).
    log :
        Log object for logging the profile operations.

    Yields
    ------
    profile : `cProfile.Profile` or `None`
        If profiling is enabled, the context manager returns the
        `cProfile.Profile` object (otherwise it returns `None`)

    Notes
    -----
    The returned profile object allows additional control
    over profiling.  You can obtain this using the ``as`` clause, e.g.:

    .. code-block:: py

        with profile(filename) as prof:
            runYourCodeHere()

    The output cumulative profile can be printed with a command-line like:

    .. code-block:: bash

        python -c 'import pstats; pstats.Stats("<filename>").\
                   sort_stats("cumtime").print_stats(30)'
    """
    if not filename:
        # Nothing to do
        yield
        return
    from cProfile import Profile

    prof = Profile()
    if log is not None:
        log.info("Enabling cProfile profiling")
    prof.enable()
    yield prof
    prof.disable()
    prof.dump_stats(filename)
    if log is not None:
        log.info("cProfile stats written to %s" % filename)


def printTable(rows: list[tuple], header: tuple | None) -> None:
    """Nice formatting of 2-column table.

    Parameters
    ----------
    rows : `list` of `tuple`
        Each item in the list is a 2-tuple containg left and righ column values
    header: `tuple` or `None`
        If `None` then table header are not prined, otherwise it's a 2-tuple
        with column headings.
    """
    if not rows:
        return
    width = max(len(x[0]) for x in rows)
    if header:
        width = max(width, len(header[0]))
        print(header[0].ljust(width), header[1])
        print("".ljust(width, "-"), "".ljust(len(header[1]), "-"))
    for col1, col2 in rows:
        print(col1.ljust(width), col2)


def filterTasks(pipeline: Pipeline, name: str | None) -> list[TaskDef]:
    """Find list of tasks matching given name.

    For matching task either task label or task name after last dot should
    be identical to `name`. If task label is non-empty then task name is not
    checked.

    Parameters
    ----------
    pipeline : `Pipeline`
        Pipeline to examine.
    name : `str` or `None`
        If empty or `None` then all tasks are returned.

    Returns
    -------
    tasks : `list` [ `lsst.pipe.base.TaskDef`]
        List of `~lsst.pipe.base.TaskDef` instances that match.
    """
    if not name:
        return list(pipeline.toExpandedPipeline())
    tasks = []
    for taskDef in pipeline.toExpandedPipeline():
        if taskDef.label:
            if taskDef.label == name:
                tasks.append(taskDef)
        elif taskDef.taskName.split(".")[-1] == name:
            tasks.append(taskDef)
    return tasks


def subTaskIter(config: pexConfig.Config) -> Iterator[tuple[str, str]]:
    """Recursively generates subtask names.

    Parameters
    ----------
    config : `lsst.pex.config.Config`
        Configuration of the task

    Returns
    -------
    names : `collections.abc.Iterator` [ `tuple` [ `str`, `str` ] ]
        Iterator which returns tuples of (configFieldPath, taskName).
    """
    for fieldName, field in sorted(config.items()):
        if hasattr(field, "value") and hasattr(field, "target"):
            subConfig = field.value
            if isinstance(subConfig, pexConfig.Config):
                try:
                    taskName = f"{field.target.__module__}.{field.target.__name__}"
                except Exception:
                    taskName = repr(field.target)
                yield fieldName, taskName
                for subFieldName, taskName in subTaskIter(subConfig):
                    yield fieldName + "." + subFieldName, taskName



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for ctrl_mpexec CLI update-graph-run subcommand."""

import os
import unittest

from lsst.ctrl.mpexec.cli.pipetask import cli as pipetask_cli
from lsst.daf.butler.cli.utils import LogCliRunner, clickResultMsg
from lsst.daf.butler.tests.utils import makeTestTempDir, removeTestTempDir
from lsst.pipe.base import QuantumGraph
from lsst.pipe.base.tests.simpleQGraph import makeSimpleQGraph
from lsst.pipe.base.tests.util import check_output_run

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class UpdateGraphRunTest(unittest.TestCase):
    """Test executing "pipetask update-graph-run" commands."""

    instrument = "lsst.pipe.base.tests.simpleQGraph.SimpleInstrument"

    def setUp(self) -> None:
        self.runner = LogCliRunner()
        self.root = makeTestTempDir(TESTDIR)

    def tearDown(self) -> None:
        removeTestTempDir(self.root)

    def test_update(self):
        """Test for updating output run in a graph."""
        nQuanta = 3
        metadata = {"output_run": "run"}
        _, qgraph = makeSimpleQGraph(
            nQuanta,
            run="run",
            root=self.root,
            instrument=self.instrument,
            metadata=metadata,
        )
        self.assertEqual(check_output_run(qgraph, "run"), [])

        old_path = os.path.join(self.root, "graph.qgraph")
        qgraph.saveUri(old_path)

        new_path = os.path.join(self.root, "graph-updated.qgraph")
        result = self.runner.invoke(
            pipetask_cli,
            ["update-graph-run", old_path, "new-run", new_path],
            input="no",
        )
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        updated_graph = QuantumGraph.loadUri(new_path)
        self.assertEqual(check_output_run(updated_graph, "new-run"), [])
        assert updated_graph.metadata is not None
        self.assertEqual(updated_graph.metadata["output_run"], "new-run")
        self.assertEqual(updated_graph.graphID, qgraph.graphID)

        # Check that we can turn off metadata updates.
        result = self.runner.invoke(
            pipetask_cli,
            ["update-graph-run", "--metadata-run-key=''", old_path, "new-run2", new_path],
            input="no",
        )
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        updated_graph = QuantumGraph.loadUri(new_path)
        self.assertEqual(check_output_run(updated_graph, "new-run2"), [])
        assert updated_graph.metadata is not None
        self.assertEqual(updated_graph.metadata["output_run"], "run")

        # Now check that we can make new graph ID.
        result = self.runner.invoke(
            pipetask_cli,
            ["update-graph-run", "--update-graph-id", old_path, "new-run3", new_path],
            input="no",
        )
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        updated_graph = QuantumGraph.loadUri(new_path)
        self.assertEqual(check_output_run(updated_graph, "new-run3"), [])
        self.assertNotEqual(updated_graph.graphID, qgraph.graphID)


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("SimplePipelineExecutor",)

from collections.abc import Iterable, Iterator, Mapping
from typing import Any

from lsst.daf.butler import Butler, CollectionType, Quantum
from lsst.pex.config import Config
from lsst.pipe.base import (
    ExecutionResources,
    GraphBuilder,
    Instrument,
    Pipeline,
    PipelineTask,
    QuantumGraph,
    TaskDef,
)

from .preExecInit import PreExecInit
from .singleQuantumExecutor import SingleQuantumExecutor
from .taskFactory import TaskFactory


class SimplePipelineExecutor:
    """A simple, high-level executor for pipelines.

    Parameters
    ----------
    quantum_graph : `~lsst.pipe.base.QuantumGraph`
        Graph to be executed.
    butler : `~lsst.daf.butler.Butler`
        Object that manages all I/O.  Must be initialized with `collections`
        and `run` properties that correspond to the input and output
        collections, which must be consistent with those used to create
        ``quantum_graph``.
    resources : `~lsst.pipe.base.ExecutionResources`
        The resources available to each quantum being executed.

    Notes
    -----
    Most callers should use one of the `classmethod` factory functions
    (`from_pipeline_filename`, `from_task_class`, `from_pipeline`) instead of
    invoking the constructor directly; these guarantee that the
    `~lsst.daf.butler.Butler` and `~lsst.pipe.base.QuantumGraph` are created
    consistently.

    This class is intended primarily to support unit testing and small-scale
    integration testing of `~lsst.pipe.base.PipelineTask` classes.  It
    deliberately lacks many features present in the command-line-only
    ``pipetask`` tool in order to keep the implementation simple.  Python
    callers that need more sophistication should call lower-level tools like
    `~lsst.pipe.base.GraphBuilder`, `PreExecInit`, and `SingleQuantumExecutor`
    directly.
    """

    def __init__(
        self,
        quantum_graph: QuantumGraph,
        butler: Butler,
        resources: ExecutionResources | None = None,
    ):
        self.quantum_graph = quantum_graph
        self.butler = butler
        self.resources = resources

    @classmethod
    def prep_butler(
        cls,
        root: str,
        inputs: Iterable[str],
        output: str,
        output_run: str | None = None,
    ) -> Butler:
        """Return configured `~lsst.daf.butler.Butler`.

        Helper method for creating `~lsst.daf.butler.Butler` instances with
        collections appropriate for processing.

        Parameters
        ----------
        root : `str`
            Root of the butler data repository; must already exist, with all
            necessary input data.
        inputs : `~collections.abc.Iterable` [ `str` ]
            Collections to search for all input datasets, in search order.
        output : `str`
            Name of a new output `~lsst.daf.butler.CollectionType.CHAINED`
            collection to create that will combine both inputs and outputs.
        output_run : `str`, optional
            Name of the output `~lsst.daf.butler.CollectionType.RUN` that will
            directly hold all output datasets.  If not provided, a name will
            be created from ``output`` and a timestamp.

        Returns
        -------
        butler : `~lsst.daf.butler.Butler`
            Butler client instance compatible with all `classmethod` factories.
            Always writeable.
        """
        if output_run is None:
            output_run = f"{output}/{Instrument.makeCollectionTimestamp()}"
        # Make initial butler with no collections, since we haven't created
        # them yet.
        butler = Butler(root, writeable=True)
        butler.registry.registerCollection(output_run, CollectionType.RUN)
        butler.registry.registerCollection(output, CollectionType.CHAINED)
        collections = [output_run]
        collections.extend(inputs)
        butler.registry.setCollectionChain(output, collections)
        # Remake butler to let it infer default data IDs from collections, now
        # that those collections exist.
        return Butler(butler=butler, collections=[output], run=output_run)

    @classmethod
    def from_pipeline_filename(
        cls,
        pipeline_filename: str,
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
        resources: ExecutionResources | None = None,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from an on-disk
        pipeline YAML file.

        Parameters
        ----------
        pipeline_filename : `str`
            Name of the YAML file to load the pipeline definition from.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `~collections.abc.Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.
        resources : `~lsst.pipe.base.ExecutionResources`
            The resources available to each quantum being executed.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed
            `~lsst.pipe.base.QuantumGraph` and `~lsst.daf.butler.Butler`,
            ready for `run` to be called.
        """
        pipeline = Pipeline.fromFile(pipeline_filename)
        return cls.from_pipeline(pipeline, butler=butler, where=where, bind=bind, resources=resources)

    @classmethod
    def from_task_class(
        cls,
        task_class: type[PipelineTask],
        config: Config | None = None,
        label: str | None = None,
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
        resources: ExecutionResources | None = None,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from a pipeline
        containing a single task.

        Parameters
        ----------
        task_class : `type`
            A concrete `~lsst.pipe.base.PipelineTask` subclass.
        config : `~lsst.pex.config.Config`, optional
            Configuration for the task.  If not provided, task-level defaults
            will be used (no per-instrument overrides).
        label : `str`, optional
            Label for the task in its pipeline; defaults to
            ``task_class._DefaultName``.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `~collections.abc.Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.
        resources : `~lsst.pipe.base.ExecutionResources`
            The resources available to each quantum being executed.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed
            `~lsst.pipe.base.QuantumGraph` and `~lsst.daf.butler.Butler`,
            ready for `run` to be called.
        """
        if config is None:
            config = task_class.ConfigClass()
        if label is None:
            label = task_class._DefaultName
        if not isinstance(config, task_class.ConfigClass):
            raise TypeError(
                f"Invalid config class type: expected {task_class.ConfigClass.__name__}, "
                f"got {type(config).__name__}."
            )
        task_def = TaskDef(taskName=task_class.__name__, config=config, label=label, taskClass=task_class)
        return cls.from_pipeline([task_def], butler=butler, where=where, bind=bind, resources=resources)

    @classmethod
    def from_pipeline(
        cls,
        pipeline: Pipeline | Iterable[TaskDef],
        *,
        where: str = "",
        bind: Mapping[str, Any] | None = None,
        butler: Butler,
        resources: ExecutionResources | None = None,
        **kwargs: Any,
    ) -> SimplePipelineExecutor:
        """Create an executor by building a QuantumGraph from an in-memory
        pipeline.

        Parameters
        ----------
        pipeline : `~lsst.pipe.base.Pipeline` or \
                `~collections.abc.Iterable` [ `~lsst.pipe.base.TaskDef` ]
            A Python object describing the tasks to run, along with their
            labels and configuration.
        where : `str`, optional
            Data ID query expression that constraints the quanta generated.
        bind : `~collections.abc.Mapping`, optional
            Mapping containing literal values that should be injected into the
            ``where`` expression, keyed by the identifiers they replace.
        butler : `~lsst.daf.butler.Butler`
            Butler that manages all I/O.  `prep_butler` can be used to create
            one.
        resources : `~lsst.pipe.base.ExecutionResources`
            The resources available to each quantum being executed.

        Returns
        -------
        executor : `SimplePipelineExecutor`
            An executor instance containing the constructed
            `~lsst.pipe.base.QuantumGraph` and `~lsst.daf.butler.Butler`,
            ready for `run` to be called.
        """
        if isinstance(pipeline, Pipeline):
            pipeline = list(pipeline.toExpandedPipeline())
        else:
            pipeline = list(pipeline)
        graph_builder = GraphBuilder(butler.registry)
        assert butler.run is not None, "Butler output run collection must be defined"
        quantum_graph = graph_builder.makeGraph(
            pipeline, collections=butler.collections, run=butler.run, userQuery=where, bind=bind
        )
        return cls(quantum_graph=quantum_graph, butler=butler, resources=resources)

    def run(self, register_dataset_types: bool = False, save_versions: bool = True) -> list[Quantum]:
        """Run all the quanta in the `~lsst.pipe.base.QuantumGraph` in
        topological order.

        Use this method to run all quanta in the graph.  Use
        `as_generator` to get a generator to run the quanta one at
        a time.

        Parameters
        ----------
        register_dataset_types : `bool`, optional
            If `True`, register all output dataset types before executing any
            quanta.
        save_versions : `bool`, optional
            If `True` (default), save a package versions dataset.

        Returns
        -------
        quanta : `list` [ `~lsst.daf.butler.Quantum` ]
            Executed quanta.

        Notes
        -----
        A topological ordering is not in general unique, but no other
        guarantees are made about the order in which quanta are processed.
        """
        return list(
            self.as_generator(register_dataset_types=register_dataset_types, save_versions=save_versions)
        )

    def as_generator(
        self, register_dataset_types: bool = False, save_versions: bool = True
    ) -> Iterator[Quantum]:
        """Yield quanta in the `~lsst.pipe.base.QuantumGraph` in topological
        order.

        These quanta will be run as the returned generator is iterated
        over.  Use this method to run the quanta one at a time.
        Use `run` to run all quanta in the graph.

        Parameters
        ----------
        register_dataset_types : `bool`, optional
            If `True`, register all output dataset types before executing any
            quanta.
        save_versions : `bool`, optional
            If `True` (default), save a package versions dataset.

        Returns
        -------
        quanta : `~collections.abc.Iterator` [ `~lsst.daf.butler.Quantum` ]
            Executed quanta.

        Notes
        -----
        Global initialization steps (see `PreExecInit`) are performed
        immediately when this method is called, but individual quanta are not
        actually executed until the returned iterator is iterated over.

        A topological ordering is not in general unique, but no other
        guarantees are made about the order in which quanta are processed.
        """
        task_factory = TaskFactory()
        pre_exec_init = PreExecInit(self.butler, task_factory)
        pre_exec_init.initialize(
            graph=self.quantum_graph, registerDatasetTypes=register_dataset_types, saveVersions=save_versions
        )
        single_quantum_executor = SingleQuantumExecutor(self.butler, task_factory, resources=self.resources)
        # Important that this returns a generator expression rather than being
        # a generator itself; that is what makes the PreExecInit stuff above
        # happen immediately instead of when the first quanta is executed,
        # which might be useful for callers who want to check the state of the
        # repo in between.
        return (single_quantum_executor.execute(qnode.taskDef, qnode.quantum) for qnode in self.quantum_graph)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from abc import ABC, abstractmethod
from collections.abc import Callable

import click


class ConfirmableResult(ABC):
    """Interface for a results class that can be used by the `confirm`
    function.
    """

    @abstractmethod
    def describe(self, will: bool) -> str:
        """Get a message describing what will be or was done. This is printed
        just before "Continue?" on the CLI, if confirming, or just before
        "Done." if confirmation is being skipped.

        Parameters
        ----------
        will : bool
            True if confirmation is being requested, False if --no-confirm was
            used, and the action is completed.
        """
        pass

    @abstractmethod
    def on_confirmation(self) -> None:
        """Perform the action that was returned from `describe`.

        This is Called just after the user has confirmed (if needed).
        """
        pass

    @property
    @abstractmethod
    def failed(self) -> bool:
        """Query if there was a failure preparing the ConfirmableResult,
        before `on_confirmation` is called.
        """
        pass

    @property
    @abstractmethod
    def describe_failure(self) -> str:
        """Get a message describing the failure.

        This is used as the message when raising a `~click.ClickException` to
        stop  with exit code 1.
        """
        pass

    @property
    @abstractmethod
    def can_continue(self) -> bool:
        """Query if the ConfirmableResult can continue. Returns `False` if
        there is no work to be done.
        """
        pass


def confirm(script_func: Callable[[], ConfirmableResult], confirm: bool) -> ConfirmableResult:
    """Prompt user to continue."""
    result = script_func()
    if result.failed:
        raise click.ClickException(result.describe_failure)
    if not result.can_continue:
        print(result.describe(will=True))
        return result
    if confirm:
        print(result.describe(will=True))
        if result.can_continue:
            do_continue = click.confirm("Continue?", default=False)
    else:
        do_continue = True
    if not do_continue:
        print("Aborted.")
    else:
        result.on_confirmation()
        if not confirm:
            print(result.describe(will=False))
        else:
            print("Done.")
    return result



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import annotations

__all__ = [
    "SeparablePipelineExecutor",
]


import datetime
import getpass
import logging
import math
import multiprocessing
from collections.abc import Iterable, Mapping
from typing import Any, Protocol

import lsst.pipe.base
import lsst.resources
from lsst.daf.butler import Butler

from .mpGraphExecutor import MPGraphExecutor
from .preExecInit import PreExecInit
from .quantumGraphExecutor import QuantumGraphExecutor
from .singleQuantumExecutor import SingleQuantumExecutor
from .taskFactory import TaskFactory

_LOG = logging.getLogger(__name__)


# Only way to keep black, flake8, and mypy all happy
_dqc = lsst.pipe.base._datasetQueryConstraints


class _GraphBuilderLike(Protocol):
    def makeGraph(
        self,
        pipeline: lsst.pipe.base.Pipeline | Iterable[lsst.pipe.base.pipeline.TaskDef],
        collections: Any,
        run: str,
        userQuery: str | None,
        datasetQueryConstraint: _dqc.DatasetQueryConstraintVariant = _dqc._ALL,
        metadata: Mapping[str, Any] | None = None,
        bind: Mapping[str, Any] | None = None,
    ) -> lsst.pipe.base.QuantumGraph:
        pass


class SeparablePipelineExecutor:
    """An executor that allows each step of pipeline execution to be
    run independently.

    The executor can run any or all of the following steps:

        * pre-execution initialization
        * pipeline building
        * quantum graph generation
        * quantum graph execution

    Any of these steps can also be handed off to external code without
    compromising the remaining ones.

    Parameters
    ----------
    butler : `lsst.daf.butler.Butler`
        A Butler whose ``collections`` and ``run`` attributes contain the input
        and output collections to use for processing.
    clobber_output : `bool`, optional
        If set, the pipeline execution overwrites existing output files.
        Otherwise, any conflict between existing and new outputs is an error.
    skip_existing_in : iterable [`str`], optional
        If not empty, the pipeline execution searches the listed collections
        for existing outputs, and skips any quanta that have run to completion
        (or have no work to do). Otherwise, all tasks are attempted (subject
        to ``clobber_output``).
    task_factory : `lsst.pipe.base.TaskFactory`, optional
        A custom task factory for use in pre-execution and execution. By
        default, a new instance of `lsst.ctrl.mpexec.TaskFactory` is used.
    resources : `~lsst.pipe.base.ExecutionResources`
        The resources available to each quantum being executed.
    """

    def __init__(
        self,
        butler: Butler,
        clobber_output: bool = False,
        skip_existing_in: Iterable[str] | None = None,
        task_factory: lsst.pipe.base.TaskFactory | None = None,
        resources: lsst.pipe.base.ExecutionResources | None = None,
    ):
        self._butler = Butler(butler=butler, collections=butler.collections, run=butler.run)
        if not self._butler.collections:
            raise ValueError("Butler must specify input collections for pipeline.")
        if not self._butler.run:
            raise ValueError("Butler must specify output run for pipeline.")

        self._clobber_output = clobber_output
        self._skip_existing_in = list(skip_existing_in) if skip_existing_in else []

        self._task_factory = task_factory if task_factory else TaskFactory()
        self.resources = resources

    def pre_execute_qgraph(
        self,
        graph: lsst.pipe.base.QuantumGraph,
        register_dataset_types: bool = False,
        save_init_outputs: bool = True,
        save_versions: bool = True,
    ) -> None:
        """Run pre-execution initialization.

        This method will be deprecated after DM-38041, to be replaced with a
        method that takes either a `~lsst.pipe.base.Pipeline` or a
        ``ResolvedPipelineGraph`` instead of a `~lsst.pipe.base.QuantumGraph`.

        Parameters
        ----------
        graph : `lsst.pipe.base.QuantumGraph`
            The quantum graph defining the pipeline and datasets to
            be initialized.
        register_dataset_types : `bool`, optional
            If `True`, register all output dataset types from the pipeline
            represented by ``graph``.
        save_init_outputs : `bool`, optional
            If `True`, create init-output datasets in this object's output run.
        save_versions : `bool`, optional
            If `True`, save a package versions dataset.
        """
        pre_exec_init = PreExecInit(self._butler, self._task_factory, extendRun=self._clobber_output)
        pre_exec_init.initialize(
            graph=graph,
            saveInitOutputs=save_init_outputs,
            registerDatasetTypes=register_dataset_types,
            saveVersions=save_versions,
        )

    def make_pipeline(self, pipeline_uri: str | lsst.resources.ResourcePath) -> lsst.pipe.base.Pipeline:
        """Build a pipeline from pipeline and configuration information.

        Parameters
        ----------
        pipeline_uri : `str` or `lsst.resources.ResourcePath`
            URI to a file containing a pipeline definition. A URI fragment may
            be used to specify a subset of the pipeline, as described in
            :ref:`pipeline-running-intro`.

        Returns
        -------
        pipeline : `lsst.pipe.base.Pipeline`
            The fully-built pipeline.
        """
        return lsst.pipe.base.Pipeline.from_uri(pipeline_uri)

    def make_quantum_graph(
        self, pipeline: lsst.pipe.base.Pipeline, where: str = "", builder: _GraphBuilderLike | None = None
    ) -> lsst.pipe.base.QuantumGraph:
        """Build a quantum graph from a pipeline and input datasets.

        Parameters
        ----------
        pipeline : `lsst.pipe.base.Pipeline`
            The pipeline for which to generate a quantum graph.
        where : `str`, optional
            A data ID query that constrains the quanta generated.
        builder : `lsst.pipe.base.GraphBuilder`-like, optional
            A graph builder that implements a
            `~lsst.pipe.base.GraphBuilder.makeGraph` method. By default, a new
            instance of `lsst.pipe.base.GraphBuilder` is used.

        Returns
        -------
        graph : `lsst.pipe.base.QuantumGraph`
            The quantum graph for ``pipeline`` as run on the datasets
            identified by ``where``.

        Notes
        -----
        This method does no special handling of empty quantum graphs. If
        needed, clients can use `len` to test if the returned graph is empty.
        """
        if not builder:
            builder = lsst.pipe.base.GraphBuilder(
                self._butler.registry,
                skipExistingIn=self._skip_existing_in,
                clobberOutputs=self._clobber_output,
            )

        metadata = {
            "input": self._butler.collections,
            "output_run": self._butler.run,
            "skip_existing_in": self._skip_existing_in,
            "skip_existing": bool(self._skip_existing_in),
            "data_query": where,
            "user": getpass.getuser(),
            "time": str(datetime.datetime.now()),
        }
        assert self._butler.run is not None, "Butler output run collection must be defined"
        graph = builder.makeGraph(
            pipeline,
            self._butler.collections,
            self._butler.run,
            userQuery=where,
            metadata=metadata,
        )
        _LOG.info(
            "QuantumGraph contains %d quanta for %d tasks, graph ID: %r",
            len(graph),
            len(graph.taskGraph),
            graph.graphID,
        )
        return graph

    def run_pipeline(
        self,
        graph: lsst.pipe.base.QuantumGraph,
        fail_fast: bool = False,
        graph_executor: QuantumGraphExecutor | None = None,
    ) -> None:
        """Run a pipeline in the form of a prepared quantum graph.

        Pre-execution initialization must have already been run;
        see `pre_execute_qgraph`.

        Parameters
        ----------
        graph : `lsst.pipe.base.QuantumGraph`
            The pipeline and datasets to execute.
        fail_fast : `bool`, optional
            If `True`, abort all (parallel) execution if any task fails (only
            used with the default graph executor).
        graph_executor : `lsst.ctrl.mpexec.QuantumGraphExecutor`, optional
            A custom graph executor. By default, a new instance of
            `lsst.ctrl.mpexec.MPGraphExecutor` is used.
        """
        if not graph_executor:
            quantum_executor = SingleQuantumExecutor(
                self._butler,
                self._task_factory,
                skipExistingIn=self._skip_existing_in,
                clobberOutputs=self._clobber_output,
                resources=self.resources,
            )
            graph_executor = MPGraphExecutor(
                numProc=math.ceil(0.8 * multiprocessing.cpu_count()),
                timeout=2_592_000.0,  # In practice, timeout is never helpful; set to 30 days.
                quantumExecutor=quantum_executor,
                failFast=fail_fast,
            )
            # Have to reset connection pool to avoid sharing connections with
            # forked processes.
            self._butler.registry.resetConnectionPool()

        graph_executor.execute(graph)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import click
from lsst.daf.butler.cli.butler import LoaderCLI
from lsst.daf.butler.cli.opt import (
    log_file_option,
    log_label_option,
    log_level_option,
    log_tty_option,
    long_log_option,
)


class PipetaskCLI(LoaderCLI):
    """Pipetask-specific implementation of command-line loader."""

    localCmdPkg = "lsst.ctrl.mpexec.cli.cmd"


@click.command(cls=PipetaskCLI, context_settings=dict(help_option_names=["-h", "--help"]))
@log_level_option()
@long_log_option()
@log_file_option()
@log_tty_option()
@log_label_option()
def cli(log_level, long_log, log_file, log_tty, log_label) -> None:  # type: ignore
    """Implement pipetask command line."""
    # log_level is handled by get_command and list_commands, and is called in
    # one of those functions before this is called.
    pass


def main() -> None:
    """Run pipetask command-line."""
    return cli()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from lsst.pipe.base import QuantumGraph
from lsst.resources import ResourcePathExpression


def update_graph_run(
    input_graph: ResourcePathExpression,
    run: str,
    output_graph: ResourcePathExpression,
    metadata_run_key: str,
    update_graph_id: bool,
) -> None:
    """Update quantum graph with new output run name and dataset IDs and save
    updated graph to a file.

    Parameters
    ----------
    input_graph : `~lsst.resources.ResourcePathExpression`
        Location of a file with existing quantum graph.
    run : `str`
        Collection name, if collection exists it must be of ``RUN`` type.
    output_graph : `~lsst.resources.ResourcePathExpression`
        Location to store updated quantum graph.
    update_graph_id : `bool`
        If `True` then also update graph ID with a new unique value.
    """
    qgraph = QuantumGraph.loadUri(input_graph)
    key = metadata_run_key if metadata_run_key else None
    qgraph.updateRun(run, metadata_key=key, update_graph_id=update_graph_id)
    qgraph.saveUri(output_graph)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simple unit test for cmdLineFwk module.
"""

import faulthandler
import logging
import os
import signal
import sys
import time
import unittest
import warnings
from multiprocessing import Manager

import networkx as nx
import psutil
from lsst.ctrl.mpexec import (
    ExecutionStatus,
    MPGraphExecutor,
    MPGraphExecutorError,
    MPTimeoutError,
    QuantumExecutor,
    QuantumReport,
    SingleQuantumExecutor,
)
from lsst.ctrl.mpexec.execFixupDataId import ExecFixupDataId
from lsst.daf.butler.tests.utils import makeTestTempDir, removeTestTempDir
from lsst.pipe.base import NodeId
from lsst.pipe.base.tests.simpleQGraph import AddTaskFactoryMock, makeSimpleQGraph

logging.basicConfig(level=logging.DEBUG)

_LOG = logging.getLogger(__name__)

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class QuantumExecutorMock(QuantumExecutor):
    """Mock class for QuantumExecutor"""

    def __init__(self, mp=False):
        self.quanta = []
        if mp:
            # in multiprocess mode use shared list
            manager = Manager()
            self.quanta = manager.list()
        self.report = None
        self._execute_called = False

    def execute(self, taskDef, quantum):
        _LOG.debug("QuantumExecutorMock.execute: taskDef=%s dataId=%s", taskDef, quantum.dataId)
        self._execute_called = True
        if taskDef.taskClass:
            try:
                # only works for one of the TaskMock classes below
                taskDef.taskClass().runQuantum()
                self.report = QuantumReport(dataId=quantum.dataId, taskLabel=taskDef.label)
            except Exception as exc:
                self.report = QuantumReport.from_exception(
                    exception=exc,
                    dataId=quantum.dataId,
                    taskLabel=taskDef.label,
                )
                raise
        self.quanta.append(quantum)
        return quantum

    def getReport(self):
        if not self._execute_called:
            raise RuntimeError("getReport called before execute")
        return self.report

    def getDataIds(self, field):
        """Return values for dataId field for each visited quanta."""
        return [quantum.dataId[field] for quantum in self.quanta]


class QuantumMock:
    """Mock equivalent of a `~lsst.daf.butler.Quantum`."""

    def __init__(self, dataId):
        self.dataId = dataId

    def __eq__(self, other):
        return self.dataId == other.dataId

    def __hash__(self):
        # dict.__eq__ is order-insensitive
        return hash(tuple(sorted(kv for kv in self.dataId.items())))


class QuantumIterDataMock:
    """Simple class to mock QuantumIterData."""

    def __init__(self, index, taskDef, **dataId):
        self.index = index
        self.taskDef = taskDef
        self.quantum = QuantumMock(dataId)
        self.dependencies = set()
        self.nodeId = NodeId(index, "DummyBuildString")


class QuantumGraphMock:
    """Mock for quantum graph."""

    def __init__(self, qdata):
        self._graph = nx.DiGraph()
        previous = qdata[0]
        for node in qdata[1:]:
            self._graph.add_edge(previous, node)
            previous = node

    def __iter__(self):
        yield from nx.topological_sort(self._graph)

    def __len__(self):
        return len(self._graph)

    def findTaskDefByLabel(self, label):
        for q in self:
            if q.taskDef.label == label:
                return q.taskDef

    def getQuantaForTask(self, taskDef):
        nodes = self.getNodesForTask(taskDef)
        return {q.quantum for q in nodes}

    def getNodesForTask(self, taskDef):
        quanta = set()
        for q in self:
            if q.taskDef == taskDef:
                quanta.add(q)
        return quanta

    @property
    def graph(self):
        return self._graph

    def findCycle(self):
        return []

    def determineInputsToQuantumNode(self, node):
        result = set()
        for n in node.dependencies:
            for otherNode in self:
                if otherNode.index == n:
                    result.add(otherNode)
        return result


class TaskMockMP:
    """Simple mock class for task supporting multiprocessing."""

    canMultiprocess = True

    def runQuantum(self):
        _LOG.debug("TaskMockMP.runQuantum")
        pass


class TaskMockFail:
    """Simple mock class for task which fails."""

    canMultiprocess = True

    def runQuantum(self):
        _LOG.debug("TaskMockFail.runQuantum")
        raise ValueError("expected failure")


class TaskMockCrash:
    """Simple mock class for task which fails."""

    canMultiprocess = True

    def runQuantum(self):
        _LOG.debug("TaskMockCrash.runQuantum")
        # Disable fault handler to suppress long scary traceback.
        faulthandler.disable()
        signal.raise_signal(signal.SIGILL)


class TaskMockLongSleep:
    """Simple mock class for task which "runs" for very long time."""

    canMultiprocess = True

    def runQuantum(self):
        _LOG.debug("TaskMockLongSleep.runQuantum")
        time.sleep(100.0)


class TaskMockNoMP:
    """Simple mock class for task not supporting multiprocessing."""

    canMultiprocess = False


class TaskDefMock:
    """Simple mock class for task definition in a pipeline."""

    def __init__(self, taskName="Task", config=None, taskClass=TaskMockMP, label="task1"):
        self.taskName = taskName
        self.config = config
        self.taskClass = taskClass
        self.label = label

    def __str__(self):
        return f"TaskDefMock(taskName={self.taskName}, taskClass={self.taskClass.__name__})"


def _count_status(report, status):
    """Count number of quanta witha a given status."""
    return len([qrep for qrep in report.quantaReports if qrep.status is status])


class MPGraphExecutorTestCase(unittest.TestCase):
    """A test case for MPGraphExecutor class"""

    def test_mpexec_nomp(self):
        """Make simple graph and execute"""
        taskDef = TaskDefMock()
        qgraph = QuantumGraphMock(
            [QuantumIterDataMock(index=i, taskDef=taskDef, detector=i) for i in range(3)]
        )

        # run in single-process mode
        qexec = QuantumExecutorMock()
        mpexec = MPGraphExecutor(numProc=1, timeout=100, quantumExecutor=qexec)
        mpexec.execute(qgraph)
        self.assertEqual(qexec.getDataIds("detector"), [0, 1, 2])
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.SUCCESS)
        self.assertIsNone(report.exitCode)
        self.assertIsNone(report.exceptionInfo)
        self.assertEqual(len(report.quantaReports), 3)
        self.assertTrue(all(qrep.status == ExecutionStatus.SUCCESS for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exitCode is None for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))
        self.assertTrue(all(qrep.taskLabel == "task1" for qrep in report.quantaReports))

    def test_mpexec_mp(self):
        """Make simple graph and execute"""
        taskDef = TaskDefMock()
        qgraph = QuantumGraphMock(
            [QuantumIterDataMock(index=i, taskDef=taskDef, detector=i) for i in range(3)]
        )

        methods = ["spawn"]
        if sys.platform == "linux":
            methods.append("fork")
            methods.append("forkserver")

        for method in methods:
            with self.subTest(startMethod=method):
                # Run in multi-process mode, the order of results is not
                # defined.
                qexec = QuantumExecutorMock(mp=True)
                mpexec = MPGraphExecutor(numProc=3, timeout=100, quantumExecutor=qexec, startMethod=method)
                mpexec.execute(qgraph)
                self.assertCountEqual(qexec.getDataIds("detector"), [0, 1, 2])
                report = mpexec.getReport()
                self.assertEqual(report.status, ExecutionStatus.SUCCESS)
                self.assertIsNone(report.exitCode)
                self.assertIsNone(report.exceptionInfo)
                self.assertEqual(len(report.quantaReports), 3)
                self.assertTrue(all(qrep.status == ExecutionStatus.SUCCESS for qrep in report.quantaReports))
                self.assertTrue(all(qrep.exitCode == 0 for qrep in report.quantaReports))
                self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))
                self.assertTrue(all(qrep.taskLabel == "task1" for qrep in report.quantaReports))

    def test_mpexec_nompsupport(self):
        """Try to run MP for task that has no MP support which should fail"""
        taskDef = TaskDefMock(taskClass=TaskMockNoMP)
        qgraph = QuantumGraphMock(
            [QuantumIterDataMock(index=i, taskDef=taskDef, detector=i) for i in range(3)]
        )

        # run in multi-process mode
        qexec = QuantumExecutorMock()
        mpexec = MPGraphExecutor(numProc=3, timeout=100, quantumExecutor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "Task Task does not support multiprocessing"):
            mpexec.execute(qgraph)

    def test_mpexec_fixup(self):
        """Make simple graph and execute, add dependencies by executing fixup
        code.
        """
        taskDef = TaskDefMock()

        for reverse in (False, True):
            qgraph = QuantumGraphMock(
                [QuantumIterDataMock(index=i, taskDef=taskDef, detector=i) for i in range(3)]
            )

            qexec = QuantumExecutorMock()
            fixup = ExecFixupDataId("task1", "detector", reverse=reverse)
            mpexec = MPGraphExecutor(numProc=1, timeout=100, quantumExecutor=qexec, executionGraphFixup=fixup)
            mpexec.execute(qgraph)

            expected = [0, 1, 2]
            if reverse:
                expected = list(reversed(expected))
            self.assertEqual(qexec.getDataIds("detector"), expected)

    def test_mpexec_timeout(self):
        """Fail due to timeout"""
        taskDef = TaskDefMock()
        taskDefSleep = TaskDefMock(taskClass=TaskMockLongSleep)
        qgraph = QuantumGraphMock(
            [
                QuantumIterDataMock(index=0, taskDef=taskDef, detector=0),
                QuantumIterDataMock(index=1, taskDef=taskDefSleep, detector=1),
                QuantumIterDataMock(index=2, taskDef=taskDef, detector=2),
            ]
        )

        # with failFast we'll get immediate MPTimeoutError
        qexec = QuantumExecutorMock(mp=True)
        mpexec = MPGraphExecutor(numProc=3, timeout=1, quantumExecutor=qexec, failFast=True)
        with self.assertRaises(MPTimeoutError):
            mpexec.execute(qgraph)
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.TIMEOUT)
        self.assertEqual(report.exceptionInfo.className, "lsst.ctrl.mpexec.mpGraphExecutor.MPTimeoutError")
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.TIMEOUT), 1)
        self.assertTrue(any(qrep.exitCode < 0 for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))

        # with failFast=False exception happens after last task finishes
        qexec = QuantumExecutorMock(mp=True)
        mpexec = MPGraphExecutor(numProc=3, timeout=3, quantumExecutor=qexec, failFast=False)
        with self.assertRaises(MPTimeoutError):
            mpexec.execute(qgraph)
        # We expect two tasks (0 and 2) to finish successfully and one task to
        # timeout. Unfortunately on busy CPU there is no guarantee that tasks
        # finish on time, so expect more timeouts and issue a warning.
        detectorIds = set(qexec.getDataIds("detector"))
        self.assertLess(len(detectorIds), 3)
        if detectorIds != {0, 2}:
            warnings.warn(f"Possibly timed out tasks, expected [0, 2], received {detectorIds}")
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.TIMEOUT)
        self.assertEqual(report.exceptionInfo.className, "lsst.ctrl.mpexec.mpGraphExecutor.MPTimeoutError")
        self.assertGreater(len(report.quantaReports), 0)
        self.assertGreater(_count_status(report, ExecutionStatus.TIMEOUT), 0)
        self.assertTrue(any(qrep.exitCode < 0 for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))

    def test_mpexec_failure(self):
        """Failure in one task should not stop other tasks"""
        taskDef = TaskDefMock()
        taskDefFail = TaskDefMock(taskClass=TaskMockFail)
        qgraph = QuantumGraphMock(
            [
                QuantumIterDataMock(index=0, taskDef=taskDef, detector=0),
                QuantumIterDataMock(index=1, taskDef=taskDefFail, detector=1),
                QuantumIterDataMock(index=2, taskDef=taskDef, detector=2),
            ]
        )

        qexec = QuantumExecutorMock(mp=True)
        mpexec = MPGraphExecutor(numProc=3, timeout=100, quantumExecutor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "One or more tasks failed"):
            mpexec.execute(qgraph)
        self.assertCountEqual(qexec.getDataIds("detector"), [0, 2])
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.ctrl.mpexec.mpGraphExecutor.MPGraphExecutorError"
        )
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertEqual(_count_status(report, ExecutionStatus.SUCCESS), 2)
        self.assertTrue(any(qrep.exitCode > 0 for qrep in report.quantaReports))
        self.assertTrue(any(qrep.exceptionInfo is not None for qrep in report.quantaReports))

    def test_mpexec_failure_dep(self):
        """Failure in one task should skip dependents"""
        taskDef = TaskDefMock()
        taskDefFail = TaskDefMock(taskClass=TaskMockFail)
        qdata = [
            QuantumIterDataMock(index=0, taskDef=taskDef, detector=0),
            QuantumIterDataMock(index=1, taskDef=taskDefFail, detector=1),
            QuantumIterDataMock(index=2, taskDef=taskDef, detector=2),
            QuantumIterDataMock(index=3, taskDef=taskDef, detector=3),
            QuantumIterDataMock(index=4, taskDef=taskDef, detector=4),
        ]
        qdata[2].dependencies.add(1)
        qdata[4].dependencies.add(3)
        qdata[4].dependencies.add(2)

        qgraph = QuantumGraphMock(qdata)

        qexec = QuantumExecutorMock(mp=True)
        mpexec = MPGraphExecutor(numProc=3, timeout=100, quantumExecutor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "One or more tasks failed"):
            mpexec.execute(qgraph)
        self.assertCountEqual(qexec.getDataIds("detector"), [0, 3])
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.ctrl.mpexec.mpGraphExecutor.MPGraphExecutorError"
        )
        # Dependencies of failed tasks do not appear in quantaReports
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertEqual(_count_status(report, ExecutionStatus.SUCCESS), 2)
        self.assertEqual(_count_status(report, ExecutionStatus.SKIPPED), 2)
        self.assertTrue(any(qrep.exitCode > 0 for qrep in report.quantaReports))
        self.assertTrue(any(qrep.exceptionInfo is not None for qrep in report.quantaReports))

    def test_mpexec_failure_dep_nomp(self):
        """Failure in one task should skip dependents, in-process version"""
        taskDef = TaskDefMock()
        taskDefFail = TaskDefMock(taskClass=TaskMockFail)
        qdata = [
            QuantumIterDataMock(index=0, taskDef=taskDef, detector=0),
            QuantumIterDataMock(index=1, taskDef=taskDefFail, detector=1),
            QuantumIterDataMock(index=2, taskDef=taskDef, detector=2),
            QuantumIterDataMock(index=3, taskDef=taskDef, detector=3),
            QuantumIterDataMock(index=4, taskDef=taskDef, detector=4),
        ]
        qdata[2].dependencies.add(1)
        qdata[4].dependencies.add(3)
        qdata[4].dependencies.add(2)

        qgraph = QuantumGraphMock(qdata)

        qexec = QuantumExecutorMock()
        mpexec = MPGraphExecutor(numProc=1, timeout=100, quantumExecutor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "One or more tasks failed"):
            mpexec.execute(qgraph)
        self.assertCountEqual(qexec.getDataIds("detector"), [0, 3])
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.ctrl.mpexec.mpGraphExecutor.MPGraphExecutorError"
        )
        # Dependencies of failed tasks do not appear in quantaReports
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertEqual(_count_status(report, ExecutionStatus.SUCCESS), 2)
        self.assertEqual(_count_status(report, ExecutionStatus.SKIPPED), 2)
        self.assertTrue(all(qrep.exitCode is None for qrep in report.quantaReports))
        self.assertTrue(any(qrep.exceptionInfo is not None for qrep in report.quantaReports))

    def test_mpexec_failure_failfast(self):
        """Fast fail stops quickly.

        Timing delay of task #3 should be sufficient to process
        failure and raise exception.
        """
        taskDef = TaskDefMock()
        taskDefFail = TaskDefMock(taskClass=TaskMockFail)
        taskDefLongSleep = TaskDefMock(taskClass=TaskMockLongSleep)
        qdata = [
            QuantumIterDataMock(index=0, taskDef=taskDef, detector=0),
            QuantumIterDataMock(index=1, taskDef=taskDefFail, detector=1),
            QuantumIterDataMock(index=2, taskDef=taskDef, detector=2),
            QuantumIterDataMock(index=3, taskDef=taskDefLongSleep, detector=3),
            QuantumIterDataMock(index=4, taskDef=taskDef, detector=4),
        ]
        qdata[1].dependencies.add(0)
        qdata[2].dependencies.add(1)
        qdata[4].dependencies.add(3)
        qdata[4].dependencies.add(2)

        qgraph = QuantumGraphMock(qdata)

        qexec = QuantumExecutorMock(mp=True)
        mpexec = MPGraphExecutor(numProc=3, timeout=100, quantumExecutor=qexec, failFast=True)
        with self.assertRaisesRegex(MPGraphExecutorError, "failed, exit code=1"):
            mpexec.execute(qgraph)
        self.assertCountEqual(qexec.getDataIds("detector"), [0])
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.ctrl.mpexec.mpGraphExecutor.MPGraphExecutorError"
        )
        # Dependencies of failed tasks do not appear in quantaReports
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertTrue(any(qrep.exitCode > 0 for qrep in report.quantaReports))
        self.assertTrue(any(qrep.exceptionInfo is not None for qrep in report.quantaReports))

    def test_mpexec_crash(self):
        """Check task crash due to signal"""
        taskDef = TaskDefMock()
        taskDefCrash = TaskDefMock(taskClass=TaskMockCrash)
        qgraph = QuantumGraphMock(
            [
                QuantumIterDataMock(index=0, taskDef=taskDef, detector=0),
                QuantumIterDataMock(index=1, taskDef=taskDefCrash, detector=1),
                QuantumIterDataMock(index=2, taskDef=taskDef, detector=2),
            ]
        )

        qexec = QuantumExecutorMock(mp=True)
        mpexec = MPGraphExecutor(numProc=3, timeout=100, quantumExecutor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "One or more tasks failed"):
            mpexec.execute(qgraph)
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.ctrl.mpexec.mpGraphExecutor.MPGraphExecutorError"
        )
        # Dependencies of failed tasks do not appear in quantaReports
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertEqual(_count_status(report, ExecutionStatus.SUCCESS), 2)
        self.assertTrue(any(qrep.exitCode == -signal.SIGILL for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))

    def test_mpexec_crash_failfast(self):
        """Check task crash due to signal with --fail-fast"""
        taskDef = TaskDefMock()
        taskDefCrash = TaskDefMock(taskClass=TaskMockCrash)
        qgraph = QuantumGraphMock(
            [
                QuantumIterDataMock(index=0, taskDef=taskDef, detector=0),
                QuantumIterDataMock(index=1, taskDef=taskDefCrash, detector=1),
                QuantumIterDataMock(index=2, taskDef=taskDef, detector=2),
            ]
        )

        qexec = QuantumExecutorMock(mp=True)
        mpexec = MPGraphExecutor(numProc=3, timeout=100, quantumExecutor=qexec, failFast=True)
        with self.assertRaisesRegex(MPGraphExecutorError, "failed, killed by signal 4 .Illegal instruction"):
            mpexec.execute(qgraph)
        report = mpexec.getReport()
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.ctrl.mpexec.mpGraphExecutor.MPGraphExecutorError"
        )
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertTrue(any(qrep.exitCode == -signal.SIGILL for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))

    def test_mpexec_num_fd(self):
        """Check that number of open files stays reasonable"""
        taskDef = TaskDefMock()
        qgraph = QuantumGraphMock(
            [QuantumIterDataMock(index=i, taskDef=taskDef, detector=i) for i in range(20)]
        )

        this_proc = psutil.Process()
        num_fds_0 = this_proc.num_fds()

        # run in multi-process mode, the order of results is not defined
        qexec = QuantumExecutorMock(mp=True)
        mpexec = MPGraphExecutor(numProc=3, timeout=100, quantumExecutor=qexec)
        mpexec.execute(qgraph)

        num_fds_1 = this_proc.num_fds()
        # They should be the same but allow small growth just in case.
        # Without DM-26728 fix the difference would be equal to number of
        # quanta (20).
        self.assertLess(num_fds_1 - num_fds_0, 5)


class SingleQuantumExecutorTestCase(unittest.TestCase):
    """Tests for SingleQuantumExecutor implementation."""

    instrument = "lsst.pipe.base.tests.simpleQGraph.SimpleInstrument"

    def setUp(self):
        self.root = makeTestTempDir(TESTDIR)

    def tearDown(self):
        removeTestTempDir(self.root)

    def test_simple_execute(self) -> None:
        """Run execute() method in simplest setup."""
        nQuanta = 1
        butler, qgraph = makeSimpleQGraph(nQuanta, root=self.root, instrument=self.instrument)

        nodes = list(qgraph)
        self.assertEqual(len(nodes), nQuanta)
        node = nodes[0]

        taskFactory = AddTaskFactoryMock()
        executor = SingleQuantumExecutor(butler, taskFactory)
        executor.execute(node.taskDef, node.quantum)
        self.assertEqual(taskFactory.countExec, 1)

        # There must be one dataset of task's output connection
        refs = list(butler.registry.queryDatasets("add_dataset1", collections=butler.run))
        self.assertEqual(len(refs), 1)

    def test_skip_existing_execute(self) -> None:
        """Run execute() method twice, with skip_existing_in."""
        nQuanta = 1
        butler, qgraph = makeSimpleQGraph(nQuanta, root=self.root, instrument=self.instrument)

        nodes = list(qgraph)
        self.assertEqual(len(nodes), nQuanta)
        node = nodes[0]

        taskFactory = AddTaskFactoryMock()
        executor = SingleQuantumExecutor(butler, taskFactory)
        executor.execute(node.taskDef, node.quantum)
        self.assertEqual(taskFactory.countExec, 1)

        refs = list(butler.registry.queryDatasets("add_dataset1", collections=butler.run))
        self.assertEqual(len(refs), 1)
        dataset_id_1 = refs[0].id

        # Re-run it with skipExistingIn, it should not run.
        assert butler.run is not None
        executor = SingleQuantumExecutor(butler, taskFactory, skipExistingIn=[butler.run])
        executor.execute(node.taskDef, node.quantum)
        self.assertEqual(taskFactory.countExec, 1)

        refs = list(butler.registry.queryDatasets("add_dataset1", collections=butler.run))
        self.assertEqual(len(refs), 1)
        dataset_id_2 = refs[0].id
        self.assertEqual(dataset_id_1, dataset_id_2)

    def test_clobber_outputs_execute(self) -> None:
        """Run execute() method twice, with clobber_outputs."""
        nQuanta = 1
        butler, qgraph = makeSimpleQGraph(nQuanta, root=self.root, instrument=self.instrument)

        nodes = list(qgraph)
        self.assertEqual(len(nodes), nQuanta)
        node = nodes[0]

        taskFactory = AddTaskFactoryMock()
        executor = SingleQuantumExecutor(butler, taskFactory)
        executor.execute(node.taskDef, node.quantum)
        self.assertEqual(taskFactory.countExec, 1)

        refs = list(butler.registry.queryDatasets("add_dataset1", collections=butler.run))
        self.assertEqual(len(refs), 1)
        dataset_id_1 = refs[0].id

        original_dataset = butler.get(refs[0])

        # Remove the dataset ourself, and replace it with something
        # different so we can check later whether it got replaced.
        butler.pruneDatasets([refs[0]], disassociate=False, unstore=True, purge=False)
        replacement = original_dataset + 10
        butler.put(replacement, refs[0])

        # Re-run it with clobberOutputs and skipExistingIn, it should not
        # clobber but should skip instead.
        assert butler.run is not None
        executor = SingleQuantumExecutor(
            butler, taskFactory, skipExistingIn=[butler.run], clobberOutputs=True
        )
        executor.execute(node.taskDef, node.quantum)
        self.assertEqual(taskFactory.countExec, 1)

        refs = list(butler.registry.queryDatasets("add_dataset1", collections=butler.run))
        self.assertEqual(len(refs), 1)
        dataset_id_2 = refs[0].id
        self.assertEqual(dataset_id_1, dataset_id_2)

        second_dataset = butler.get(refs[0])
        self.assertEqual(list(second_dataset), list(replacement))

        # Re-run it with clobberOutputs but without skipExistingIn, it should
        # clobber.
        assert butler.run is not None
        executor = SingleQuantumExecutor(butler, taskFactory, clobberOutputs=True)
        executor.execute(node.taskDef, node.quantum)
        self.assertEqual(taskFactory.countExec, 2)

        refs = list(butler.registry.queryDatasets("add_dataset1", collections=butler.run))
        self.assertEqual(len(refs), 1)
        dataset_id_3 = refs[0].id

        third_dataset = butler.get(refs[0])
        self.assertEqual(list(third_dataset), list(original_dataset))

        # No change in UUID even after replacement
        self.assertEqual(dataset_id_1, dataset_id_3)


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import tempfile
import unittest

import click
import lsst.utils.tests
from lsst.ctrl.mpexec.cli import opt, script
from lsst.ctrl.mpexec.cli.pipetask import cli as pipetaskCli
from lsst.ctrl.mpexec.showInfo import ShowInfo
from lsst.daf.butler.cli.utils import LogCliRunner, clickResultMsg
from lsst.pipe.base import Pipeline


class BuildTestCase(unittest.TestCase):
    """Test a few of the inputs to the build script function to test basic
    funcitonality.
    """

    @staticmethod
    def buildArgs(**kwargs):
        defaultArgs = dict(
            log_level=(),
            order_pipeline=False,
            pipeline=None,
            pipeline_actions=(),
            pipeline_dot=None,
            save_pipeline=None,
            show=ShowInfo([]),
        )
        defaultArgs.update(kwargs)
        return defaultArgs

    def testMakeEmptyPipeline(self):
        """Test building a pipeline with default arguments."""
        pipeline = script.build(**self.buildArgs())
        self.assertIsInstance(pipeline, Pipeline)
        self.assertEqual(len(pipeline), 0)

    def testSavePipeline(self):
        """Test pipeline serialization."""
        with tempfile.TemporaryDirectory() as tempdir:
            # make empty pipeline and store it in a file
            filename = os.path.join(tempdir, "pipeline_file.yaml")
            pipeline = script.build(**self.buildArgs(save_pipeline=filename))
            self.assertIsInstance(pipeline, Pipeline)
            self.assertTrue(os.path.isfile(filename))
            # read pipeline from a file
            pipeline = script.build(**self.buildArgs(pipeline=filename))
            self.assertIsInstance(pipeline, Pipeline)
            self.assertIsInstance(pipeline, Pipeline)
            self.assertEqual(len(pipeline), 0)

    def testShowPipeline(self):
        """Test showing the pipeline."""

        class ShowInfoCmp:
            def __init__(self, show, expectedOutput):
                self.show = show
                self.expectedOutput = expectedOutput

            def __repr__(self):
                return f"ShowInfoCmp({self.show}, {self.expectedOutput}"

        testdata = [
            ShowInfoCmp(
                "pipeline",
                """description: anonymous
tasks:
  task:
    class: lsst.pipe.base.tests.simpleQGraph.AddTask
    config:
    - addend: '100'""",
            ),
            ShowInfoCmp(
                "config",
                """### Configuration for task `task'
# Flag to enable/disable saving of log output for a task, enabled by default.
config.saveLogOutput=True

# amount to add
config.addend=100

# name for connection input
config.connections.input='add_dataset{in_tmpl}'

# name for connection output
config.connections.output='add_dataset{out_tmpl}'

# name for connection output2
config.connections.output2='add2_dataset{out_tmpl}'

# name for connection initout
config.connections.initout='add_init_output{out_tmpl}'

# Template parameter used to format corresponding field template parameter
config.connections.in_tmpl='_in'

# Template parameter used to format corresponding field template parameter
config.connections.out_tmpl='_out'""",
            ),
            # history will contain machine-specific paths, TBD how to verify
            ShowInfoCmp("history=task::addend", None),
            ShowInfoCmp("tasks", "### Subtasks for task `lsst.pipe.base.tests.simpleQGraph.AddTask'"),
        ]

        for showInfo in testdata:
            runner = LogCliRunner()
            result = runner.invoke(
                pipetaskCli,
                [
                    "build",
                    "--task",
                    "lsst.pipe.base.tests.simpleQGraph.AddTask:task",
                    "--config",
                    "task:addend=100",
                    "--show",
                    showInfo.show,
                ],
            )
            self.assertEqual(result.exit_code, 0, clickResultMsg(result))
            if showInfo.expectedOutput is not None:
                self.assertIn(showInfo.expectedOutput, result.output, msg=f"for {showInfo}")

    def testMissingOption(self):
        """Test that the build script fails if options are missing."""

        @click.command()
        @opt.pipeline_build_options()
        def cli(**kwargs):
            script.build(**kwargs)

        runner = click.testing.CliRunner()
        result = runner.invoke(cli)
        # The cli call should fail, because script.build takes more options
        # than are defined by pipeline_build_options.
        self.assertNotEqual(result.exit_code, 0)


class QgraphTestCase(unittest.TestCase):
    """Test pipetask qgraph command-line."""

    def testMissingOption(self):
        """Test that if options for the qgraph script are missing that it
        fails.
        """

        @click.command()
        @opt.pipeline_build_options()
        def cli(**kwargs):
            script.qgraph(**kwargs)

        runner = click.testing.CliRunner()
        result = runner.invoke(cli)
        # The cli call should fail, because qgraph.build takes more options
        # than are defined by pipeline_build_options.
        self.assertNotEqual(result.exit_code, 0)


class RunTestCase(unittest.TestCase):
    """Test pipetask run command-line."""

    def testMissingOption(self):
        """Test that if options for the run script are missing that it
        fails.
        """

        @click.command()
        @opt.pipeline_build_options()
        def cli(**kwargs):
            script.run(**kwargs)

        runner = click.testing.CliRunner()
        result = runner.invoke(cli)
        # The cli call should fail, because qgraph.run takes more options
        # than are defined by pipeline_build_options.
        self.assertNotEqual(result.exit_code, 0)


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()
